#!/usr/bin/perl -w

#############################################################################
#
# dorep3.pl - SKaMPI report generator V3.0
#
# This script produces a Postscript report file out of a SKaMPI output file.
# Please adjust the first line to match the path of your Perl executable.
#
# Authors:
#   Ralf H. Reussner  - version 1 as basis for this version
#   Gunnar Hunzelmann - refinements and extensions
#
# Copyright 1999
#
# For questions and comments, please send email to
#
# Gunnar.Hunzelmann@stud.uni-karlsruhe.de
#   or
# reussner@ira.uka.de
#
#############################################################################




########## initializations ##########

$Version     = 'V3.0';
$C_Date      = '2000';
$Script_Name = 'dorep3';

####
# initialize user-adjustable variables with default values
####

# skampi.out file (with and without directory prefix)
$skampi_data_file      = "skampi.out";
$skampi_data_file_base = "skampi.out";

# .dorep file
$dorep_opt_file = ".dorep";

# hash which maps the variation parameters to the values whose results
# and node times are shown in tables or charts
%barchart_values =
(
 "message_length"  => ["min", "max"],
 "number_of_nodes" => ["min", "max"],
 "chunks"          => ["min", "max"]
);

# directory with dorep-information files (dri)
$dri_directory     = "";
# if no dri files can be found, this gets true (no warnings for single dri's)
$dri_files_missing = 0;

# number of data lines from which on a chart is created
$minimum_data = 4;

# overwrite existing files (0 = no, 1 = yes)
$overwrite = 1;  # CHANGED

# level of quietness during execution
# WARNING (when changing): sometimes compared with < or >
$verbose    = 0;  # print all messages
$quiet      = 1;  # only print warnings and errors
$very_quiet = 2;  # print nothing (except for LaTeX messages to `dorep.log')

$quietness  = $verbose;

# level of deleting temporary files
# WARNING (when changing): sometimes compared with < or >
$delete_nothing       = 0;  # temp. files are kept
$delete_temp          = 1;  # temp. files are deleted unless an error occurs
$delete_temp_on_error = 2;  # temp. files are deleted even if an error occurs

$wastefulness = $delete_temp;

# possible files for redirecting output
@redirect = ();

# wanted end product of this script; normally `.ps' is produced, but the script
# can also stop after having produced the `.tex' or the `.dvi' file.
# this is an undocumented feature.
$to_tex = 0;
$to_dvi = 0;

# all suites desired if [suites] section does not exist in `.dorep'
# suites can nonetheless be excluded with the [exclude] section in `.dorep'
$report_all_suites = 1;

# desired suites if [suites] section exists
@reported_suites = ();

# not desired suites if [exclude] section exists
@excluded_suites = ();

# desired comparisons; format: ((title, (ski-file, ...)), ...)
@comparisons = ();

# needed suites for comparisons
@compared_suites = ();

#### eo user-adjustables ####

####
# initialize non-user-adjustable variables
####

# usage messages and options
$usage_header =
  "$Script_Name.pl $Version (C) $C_Date Gunnar Hunzelmann, Ralf Reussner\n";
$usage = "Usage: $Script_Name.pl [-b barchart-info] [-d directory] [-h] ".
  "[-m number]\n                 [-p filename] [-q | -Q] [-t | -T] [-u] ".
  "[filename]\n";
$options = "Options:\n".
  "  -b, --barchart  barchart-info   info about which node times to show\n".
  "  -d, --dri_directory  pathname   path of directory where .dri ".
  "files reside\n".
  "                                  (default: current directory)\n".
  "  -h, --help                      print this help text and quit\n".
  "  -m, --minimum_data  number      number of data from which on ".
  "charts are\n".
  "                                  created instead of tables ".
  "(default: 4)\n".
  "  -p, --output  filename          write messages in file\n".
  "  -q, --quiet                     only print warnings\n".
  "  -Q, --very_quiet                don't even print warnings\n".
  "  -t, --keep_temp                 keep temporary files\n".
  "  -T, --delete_temp_on_error      delete temporary files even ".
  "when\n".
  "                                  an error occurs\n".
  "  -u, --use_temp                  use existing temporary files\n".
  "  filename                        name of SKaMPI output file\n".
  "                                  (default: skampi.out)\n";

# known names for variation parameters (for verification of user parameters)
@known_parameters = ("message_length",
                     "number_of_nodes",
                     "chunks");

# names of units of the different variation parameters
# first elements of the lists are the unit strings in a SKaMPI output file,
# second elements are the unit strings which are to be printed in the report
%unit_hash =
(
 "message_length"  => ["Bytes", "bytes"],
 "number_of_nodes" => ["number", "nodes"],
 "chunks"          => ["pieces of work", "chunks"]
);

# hash which maps the pattern strings to abbreviations
$pattern_str_hash{"Point-to-Point"} = "p2p";
$pattern_str_hash{"Master-Worker"} = "mw";
$pattern_str_hash{"Collective"} = "col";
$pattern_str_hash{"Simple"} = "simple";

# hash which maps collective MPI routines to respective bandwidth functions;
# comp_bdw_col_sng is default if none of the following expressions is part
# of the routine name.
# there's a detailed listing at the end of this script if future suite names
# should collide with the patterns listed here.
%band_width_func =
(
 "All"             => \&comp_bdw_col_all,
 "Barrier-nodes"   => undef,
 "Commsplit-nodes" => undef,
 "col_dummy"       => undef
);

# lists of temporary files to be deleted if need be
@files_to_delete = ();
@eps_files       = ();

# list of hashes with optional machine infos for comparisons
# Now the heavy part: It's a list of references to hashes whose keys' values
# are references to lists!
@comp_info_hash_list = ();

# variables which indicate which patterns are present
$existing_p2p    = 0;
$existing_mw     = 0;
$existing_col    = 0;
$existing_simple = 0;

# hash which maps the patterns to the `existing_' variables
$pattern_existing{"Point-to-Point"} = \$existing_p2p;
$pattern_existing{"Master-Worker"}  = \$existing_mw;
$pattern_existing{"Collective"}     = \$existing_col;
$pattern_existing{"Simple"}         = \$existing_simple;

# minimum number of units for measurement lengths from which to convert to a
# multiple of 1024 units
$min_bytes = 1024;

# minimum number of node times from which to produce wider barchart images
$min_node_times = 64;

#### eo non-user-adjustables ####

########## end of initializations ##########


########## main ##########

# check the environment for any options
check_environment ();

# read option file (.dorep)
read_options ();

# read command line options
parse_parameters ();

# adjust basename of skampi.out file
$skampi_data_file_base =  $skampi_data_file;
$skampi_data_file_base =~ s/.*\/(.*)/$1/;

# look for dri files
unless (<$dri_directory*.dri>)
{
    # Print warning now and don't produce warnings for every single missing
    # dri file later.
    if ($dri_directory eq '')
    {
        mywarn("No dri-files can be found in the current directory.\n");
    }
    else
    {
        mywarn("No dri-files can be found in directory `$dri_directory'.\n");
    }
    mywarn("The produced report will not include any descriptions for the ".
           "MPI functions.\n");

    $dri_files_missing = 1;
}

# possibly select file for redirection
while (@redirect)
{
    my $file = pop @redirect;
    if (open(REDIRECT, ">$file"))
    {
        select(REDIRECT);
        last;
    }
    else
    {
        mywarn("Cannot open file `$file' for redirection of output: $!.\n");
        if (@redirect)
        {
            mywarn("Trying file in parameter file `$dorep_opt_file'.\n");
        }
    }
}

# read the general information from the input file ...
%info_hash = read_input_file_header($skampi_data_file);

# test for given institution name
unless (check_institution($info_hash{"INSTITUTION"}))
{
    # If institution is not set, stop before creating any files
    quit("Please fill in an institution in `$skampi_data_file'");
}

# divide the input file in many files; two for each suite of measurements
rsplit($skampi_data_file);

# generate gnuplot commandfile, eps-files, and tex modules
$saved_arg = $_;  # $_ isn't recovered after while (<...>)
while (<*.ski>)
{
    my    $ski_file = $_;
    local $wide_nt  = 0;   # set to true in header2node_gpl_files if wide eps
                           # files were created

    next unless (($report_all_suites &&
                  !grep($ski_file eq $_, @excluded_suites)) ||
                 grep($ski_file eq $_, @reported_suites));

    @ms = read_header($ski_file);

    myprint("$ms[1]\n");

    my $result_files_ref = header2gpl_files(@ms);
    my $node_files_ref   = header2node_gpl_files(@ms);

    if (defined $result_files_ref)
    {
        foreach (@$result_files_ref)
        {
            system("gnuplot $_.gpl >$_.eps")
              && quit("Cannot execute gnuplot with command file `$_.gpl'",
                      __LINE__);
            push(@eps_files, "$_.eps");
        }
    }
    foreach (@$node_files_ref)
    {
        system("gnuplot $_.gpl >$_.eps")
          && quit("Cannot execute gnuplot with command file `$_.gpl'",
                  __LINE__);
        push(@eps_files, "$_.eps");
    }

    header2tex_module(@ms, $result_files_ref, $node_files_ref);
}
$_ = $saved_arg;

open(TEX_FILE, ">${skampi_data_file_base}.tex")
  || quit("Cannot open `${skampi_data_file_base}.tex': $!", __LINE__);
push(@files_to_delete, "${skampi_data_file_base}.tex") unless ($to_tex);

# ... write it to the output file
print_output_file_header();

myprint("** Including Point-to-Point measurements **\n") if (<p2p*.tmd>);
$header_printed = 0;
$saved_arg = $_;  # $_ isn't recovered after while (<...>)
while (<p2p*.tmd>)
{
    if (!$header_printed)
    {
        $header_printed = 1;
        print_p2p_header();
    }

    open(TEX_MODULE, "<$_")
      || quit("Cannot open `$_': $!", __LINE__);
    my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
    while (<TEX_MODULE>)
    {
        print(TEX_FILE $_);
    }
    $_ = $saved_arg;
    close(TEX_MODULE);
}
$_ = $saved_arg;

myprint("** Including Master-Worker measurements **\n") if (<mw*.tmd>);
$header_printed = 0;
$saved_arg = $_;  # $_ isn't recovered after while (<...>)
while (<mw*.tmd>)
{
    if (!$header_printed)
    {
        $header_printed = 1;
        print_mw_header();
    }

    open(TEX_MODULE, "<$_")
      || quit("Cannot open `$_': $!", __LINE__);
    my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
    while (<TEX_MODULE>)
    {
        print(TEX_FILE $_);
    }
    $_ = $saved_arg;
    close(TEX_MODULE);
}
$_ = $saved_arg;

myprint("** Including Collective measurements **\n") if (<col*.tmd>);
$header_printed = 0;
$saved_arg = $_;  # $_ isn't recovered after while (<...>)
while (<col*.tmd>)
{
    if (!$header_printed)
    {
        $header_printed = 1;
        print_col_header();
    }

    open(TEX_MODULE, "<$_")
      || quit("Cannot open `$_': $!", __LINE__);
    my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
    while (<TEX_MODULE>)
    {
        print(TEX_FILE $_);
    }
    $_ = $saved_arg;
    close(TEX_MODULE);
}
$_ = $saved_arg;

myprint("** Including Simple measurements **\n") if (<simple*.tmd>);
$header_printed = 0;
$saved_arg = $_;  # $_ isn't recovered after while (<...>)
while (<simple*.tmd>)
{
    if (!$header_printed)
    {
        $header_printed = 1;
        print_simple_header();
    }

    open(TEX_MODULE, "<$_")
      || quit("Cannot open `$_': $!", __LINE__);
    my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
    while (<TEX_MODULE>)
    {
        print(TEX_FILE $_);
    }
    $_ = $saved_arg;
    close(TEX_MODULE);
}
$_ = $saved_arg;

myprint("** Adding optional additional tex-modules **\n") if (<*.tma>);
$saved_arg = $_;  # $_ isn't recovered after while (<...>)
while (<*.tma>)
{
    open(TEX_MODULE_ADD,"<$_")
      || quit("Cannot open `$_': $!", __LINE__);
    myprint("Adding tex-module: $_\n");
    my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
    while (<TEX_MODULE_ADD>)
    {
        print(TEX_FILE $_);
    }
    $_ = $saved_arg;
    close(TEX_MODULE_ADD);
}
$_ = $saved_arg;

if (@comparisons)
{
    local $flag       = 0;             # header already printed?
    local $file_count = 0;             # counter for gpl files
    local $subsection = "subsection";  # lower level if more than one machine

    myprint("** Adding comparisons **\n");

    foreach (@comparisons)
    {
        my $title     = $$_[0];
        my @file_list = @{$$_[1]};

        myprint("Adding from parameter file `$dorep_opt_file':\n ".
                sprintf("$title:\n -- %s\n",
                        join("\n -- ", map(/(.+?)\.skic?$/, @file_list))));

        add_optional($title, @file_list);
    }
}

print_output_file_trailer();

close(TEX_FILE);

unless ($to_tex)
{
    @existing_files = <$skampi_data_file_base.*>;

    myprint("TeXing around...\n");
    system("latex ${skampi_data_file_base}.tex >dorep.log")
      && do
      {
          my $see_log = '';

          $see_log = ". Errors see `$skampi_data_file_base.log'"
            if (-f "$skampi_data_file_base.log");

          foreach $file (<$skampi_data_file_base.*>)
          {
              push(@files_to_delete, $file)
                if (!grep($file eq $_, @existing_files) &&
                    !($file =~ /\.dvi$/ && $to_dvi));
          }
          quit("Cannot execute latex$see_log", __LINE__);
      };
    myprint("2nd pass...\n");
    system("latex ${skampi_data_file_base}.tex >>dorep.log")
      && do
      {
          my $see_log = ". Errors see `$skampi_data_file_base.log'";
          foreach $file (<$skampi_data_file_base.*>)
          {
              push(@files_to_delete, $file)
                if (!grep($file eq $_, @existing_files) &&
                    !($file =~ /\.dvi$/ && $to_dvi));
          }
          quit("Cannot execute 2nd run of latex$see_log", __LINE__);
      };

    foreach $file (<$skampi_data_file_base.*>)
    {
        push(@files_to_delete, $file)
          if (!grep($file eq $_, @existing_files) &&
              !($file =~ /\.dvi$/ && $to_dvi));
    }

    unless ($to_dvi)
    {
        myprint("Dvipsing around...\n");
        system("dvips ${skampi_data_file_base}.dvi ".
               "-o ${skampi_data_file_base}.ps >>dorep.log")
          && quit("Cannot execute dvips", __LINE__);
    }
}

# Clean up temporary files
print(kill_temp($wastefulness));

close(REDIRECT);

# Exit script with value 0 which means "no errors"
exit(0);

########## that's all folks: end of main ##########



########## functions ##########

####
# read_input_file_header
#
# Reads the header of the `skampi.out' file and initializes the hash which
# holds the information for the main machine.
#
# PARAMS:
#   String - filename of the `skampi.out' file
# RETURNS:
#   the created hash (values are TeX-compatible)
####
sub read_input_file_header
{
    my $file_name = shift @_;
    my %info_hash;
    my $key;
    my $value;

    open(SKAMPI, "<$file_name")
      || quit("Cannot open `$file_name': $!", __LINE__);

    %info_hash = parse_run_info(*SKAMPI);

    foreach (keys(%info_hash))
    {
        $info_hash{$_} = '?' if ($info_hash{$_} eq '');

        $info_hash{$_} = convert2tex($info_hash{$_});
    }

    close(SKAMPI);

    return %info_hash;
}
#### eo read_input_file_header ####

####
# read_data
#
# Reads the result lines in a `.ski' file produced by SKaMPI and returns a
# reference to a list of these data lines.
#
# PARAMS:
#   String - filename of the `.ski' file to be read
# RETURNS:
#   list of data lines
####
sub read_data
{
    my $file_name  = shift @_;
    my @mylist     = ();
    my %suite_info = ();
    my $dimension;

    open(SKI_FILE,"<$file_name")
      || quit("Cannot open `$file_name': $!", __LINE__);

    %suite_info = parse_suite_info(*SKI_FILE);
    $dimension  = @{$suite_info{'parameters'}}
      if (defined($suite_info{'parameters'}));

    @mylist = parse_measurements($dimension, *SKI_FILE);

    close(SKI_FILE);

    return @mylist;
}
#### eo read_data ####

####
# print_output_file_header
#
# Prints the header of the resulting LaTeX file with information about the
# main machine.
#
# PARAMS:
#   none
# RETURNS:
#   nothing
####
sub print_output_file_header
{
    my $basetypes_exist = 0;

    print(TEX_FILE <<'EOT');
\nonstopmode
\documentclass[12pt, twoside,a4paper]{article}
\usepackage{doublespace}
\usepackage{moreverb}
\usepackage{epsfig}
\begin{document}
\setlength{\parindent}{1.5em}
\renewcommand{\topfraction}{0.99}
\renewcommand{\bottomfraction}{0.99}
\renewcommand{\textfraction}{0.0001}
\newcommand{\skampi}{{\sl SKaMPI\/}}
\begin{spacing}{1.2}

EOT
    if (defined($info_hash{"SKAMPIVERSION"}))
    {
        my $vers = $info_hash{"SKAMPIVERSION"};
        print(TEX_FILE "\\title{Measurements of \\skampi, Version $vers\\\\");
        print(TEX_FILE "(Special Karlsruher MPI-Benchmark)");
    }
    else
    {
        print(TEX_FILE "\\title{Measurements of \\skampi\\\\".
              "(Special Karlsruher MPI-Benchmark)");
    }

    if (defined($info_hash{"MACHINE"}))
    {
        my $mach = $info_hash{"MACHINE"};
        print(TEX_FILE "\\\\of the\\\\{$mach}");

        if (defined($info_hash{"INSTITUTION"}))
        {
            my $inst = $info_hash{"INSTITUTION"};
            print(TEX_FILE "\\\\at\\\\{$inst}");
        }
    }

    if (defined($info_hash{"USER"}))
    {
        my $user = $info_hash{"USER"};
        print(TEX_FILE "\\\\performed by\\\\{$user}\\\\");
    }

    if (defined($info_hash{"DATE"}))
    {
        my $date = $info_hash{"DATE"};
        print(TEX_FILE "\\Large on {$date}");
    }

    print(TEX_FILE "}");  # end of \title

    print(TEX_FILE "\\author{$Script_Name version $Version}\\date{\\today}");

    print(TEX_FILE <<'EOT');
\maketitle
~\\
\begin{center}
\begin{tabular}{lr}
EOT
    if (defined($info_hash{"OSNAME"}))
    {
        my $osname = $info_hash{"OSNAME"};
        print(TEX_FILE "Operating system: & {$osname}\\\\");
    }
    if (defined($info_hash{"OSRELEASE"}))
    {
        my $osrel = $info_hash{"OSRELEASE"};
        print(TEX_FILE "Release: & {$osrel}\\\\");
    }
    if (defined($info_hash{"OSVERSION"}))
    {
        my $osvers = $info_hash{"OSVERSION"};
        print(TEX_FILE "Version: & {$osvers}\\\\");
    }
    if (defined($info_hash{"HOSTNAME"}))
    {
        my $host = $info_hash{"HOSTNAME"};
        print(TEX_FILE "Hostname: & {$host}\\\\");
    }
    if (defined($info_hash{"ARCHITECTURE"}))
    {
        my $arch = $info_hash{"ARCHITECTURE"};
        print(TEX_FILE "Architecture: & {$arch}\\\\");
    }
    print(TEX_FILE "\\end{tabular}\\\\[1cm]");
    print(TEX_FILE <<'EOT');
\end{center}
\newpage
Output created by \skampi\ --- Special Karlsruher MPI-Benchmark\\
~\\R. H. Reussner\\
University of Karlsruhe\\
Department of Informatics\\
Germany\\
reussner@acm.org
\thispagestyle{empty}
\clearpage
\pagenumbering{roman}
\tableofcontents
%\listoffigures
\newpage
\pagenumbering{arabic}
\pagestyle{headings}
%/*@inskarep.include.tex*/
EOT

    if (defined($info_hash{"MACHINE"}))
    {
        print(TEX_FILE "\\section{Machine}\n");
        print(TEX_FILE $info_hash{"MACHINE"}."\n");

        if (defined($info_hash{"INSTITUTION"}))
        {
            print(TEX_FILE "at\n".$info_hash{"INSTITUTION"}."\n");
        }
    }
    if (defined($info_hash{"NODE"}))
    {
        print(TEX_FILE "\\subsection{Description of a Node}\n".
              $info_hash{"NODE"}."\n");
    }
    if (defined($info_hash{"NETWORK"}))
    {
        print(TEX_FILE "\\subsection{Network}\n".$info_hash{"NETWORK"}."\n");

        if (defined($info_hash{"NODES"}))
        {
            print(TEX_FILE "with\n".$info_hash{"NODES"}." nodes\n");
        }
    }
    if (defined($info_hash{"ABSOLUTE"}))
    {
        print(TEX_FILE "\\subsection{Results}\n");

        if ($info_hash{"ABSOLUTE"} =~ /yes/)
        {
            print(TEX_FILE "The results are absolute values. ".
                  "The overhead of \\skampi\\ is subtracted.\n");
        }
        else
        {
            print(TEX_FILE "The results are not absolute values. ".
                  "They include the overhead of \\skampi.\n");
            my $existing_patterns =
              grep($$_ == 1, @pattern_existing{keys %pattern_existing});

            if ($existing_patterns)
            {
                my $first_pattern = 1;

                print(TEX_FILE "The pattern's overhead is documented ".
                      "in section");
                print(TEX_FILE "s") if ($existing_patterns > 1);
                if ($existing_p2p)
                {
                    print(TEX_FILE ",") if (!$first_pattern);
                    $first_pattern = 0;
                    print(TEX_FILE " \\ref{p2p_dummy}");
                }
                if ($existing_mw)
                {
                    print(TEX_FILE ",") if (!$first_pattern);
                    $first_pattern = 0;
                    print(TEX_FILE " \\ref{mw_dummy}");
                }
                if ($existing_col)
                {
                    print(TEX_FILE ",") if (!$first_pattern);
                    $first_pattern = 0;
                    print(TEX_FILE " \\ref{col_dummy}");
                }
                if ($existing_simple)
                {
                    print(TEX_FILE ",") if (!$first_pattern);
                    $first_pattern = 0;
                    print(TEX_FILE " \\ref{simple_dummy}");
                }
                print(TEX_FILE ".");
            }  # patterns do exist
        }
    }

    if (defined($info_hash{"POSTPROC"}))
    {
        print(TEX_FILE "\\subsection{Postprocessing}\n");

        if ($info_hash{"POSTPROC"} !~ /times/)
        {
            print(TEX_FILE "The results have been created during".
                  " one run of \\skampi, there has been no postprocessing.");
        }
        else
        {
            print(TEX_FILE "For creating these results \\skampi\\ ran ".
                  $info_hash{"POSTPROC"});
        }
    }

    # look for defined basetypes
    if (my @bt_keys = grep(/^BASETYPE\d+$/, keys(%info_hash)))
    {
        my @basetypes = ();

        print(TEX_FILE "\\subsection{Basetypes}\nSee section ".
              "\\ref{basetypes} for the description of basetypes.\n\n");
        print(TEX_FILE "\\noindent\nUsed basetypes for this \\skampi\\ run ".
              "are:\\\\\n\n\\begin{tabular}{rl}\n".
              "\\textbf{Basetype} & \\textbf{Definition}\\\\\n");

        foreach (@bt_keys)
        {
            my ($num) = /(\d+)$/;

            # make sure that all is sorted correctly
            $basetypes[$num] = $info_hash{$_};
        }

        foreach (0..$#basetypes)
        {
            if (defined($basetypes[$_]))
            {
                print(TEX_FILE "  $_ & $basetypes[$_]\\\\\n");
            }
        }

        print(TEX_FILE "\\end{tabular}\n");
    }

    my $constants_str = '';

    if (my @extras = grep($_ eq 'BLOCKS' || $_ eq 'BLOCKSIZE' ||
                          $_ eq 'VECTORSTRIDE', keys(%info_hash)))
    {
        # these keywords appear in version 3.0 and later of SKaMPI from which
        # the output of the basetype section is needed
        $basetypes_exist = 1;

        $constants_str .= "\\begin{tabular}{|l|r|l|}\n\\hline\n";
        $constants_str .=
          "\\textbf{Constant} & \\textbf{Value} & \\textbf{Comment}\\\\\n".
            "\\hline\n";

        foreach (sort(@extras))
        {
            my $constant = 'params.'.lc();

            $constants_str .= "  $constant & $info_hash{$_} & ";

            $constants_str .= 'must be $>=$ params.blocksize'
              if ($_ eq 'VECTORSTRIDE');

            $constants_str .= "\\\\\n";
        }

        $constants_str .= "\\hline\n\\end{tabular}";
    }

    print_basetypes_info($constants_str) if ($basetypes_exist);
}
#### eo print_output_file_header ####

####
# print_basetypes_info
#
# Prints a section with descriptions of basetypes used in SKaMPI.
#
# PARAMS:
#   $ - string describing used constants like BLOCKS, VECTORSTRIDE, etc.
# RETURNS:
#   nothing
####
sub print_basetypes_info
{
    my $constants_str = shift;

    print(TEX_FILE << "EOT");
\\begin{spacing}{1.2}
\\section{User-defined Types in \\skampi}
\\subsection{Basetypes}
\\label{basetypes}
A basetype is a type which is used to define other types (such as
structs or vectors).  A basetype can be given in two ways: 
\\begin{itemize}
\\item as a string, denoting a predefined MPI datatype. Currently
\\texttt{MPI\\_CHAR}, \\texttt{MPI\\_BYTE}, \\texttt{MPI\\_INT},
\\texttt{MPI\\_LONG}, \\texttt{MPI\\_DOUBLE}, and \\texttt{MPI\\_LONG\\_DOUBLE}
are allowed.     
\\item or as a comma-separated list of tuples (see below.)
\\end{itemize}
Each tuple describes an element of the
basetype. The first entry of a tuple gives the length of the element
in bytes. The second element gives the displacement of this element in
bytes. I.e., the address where the element starts relative to the
beginning of the basetype. For example: \\texttt{(4,0)} denotes
something like \\texttt{MPI\\_DOUBLE} (on most machines); simply
four bytes at the beginning of the datatype. A basetype of two
\\texttt{MPI\\_CHAR}'s is decribed by \\texttt{(1,0),(1,1)}.  Currently
10 basetypes can be defined in \\skampi~ by the user.
%ver 2.7
\\subsection{Using Basetypes to Construct Other Types}
\\skampi~ defines several `type constructors' to define nested types
out of a user-defined basetype. These constructors are selected by a
number. Since working with user-defined types requires a certain
acquaintance with the appropriate MPI functions, we describe the
different constructors in documenting their code.

The used constants are given below:\\\\

\\noindent
$constants_str
\\begin{spacing}{1}
\\begin{verbatim}
#define SUBARRAYOFFSET 3
\\end{verbatim}
\\end{spacing}

\\subsubsection*{Number 1}
Simply \\texttt{MPI\\_BYTE}.
\\subsubsection*{Number 2}
Simply \\texttt{MPI\\_CHAR}.
\\subsubsection*{Number 3}
Simply \\texttt{MPI\\_DOUBLE}.
\\subsubsection*{Number 20}
\\sloppy
A user-defined basetype, selected by the
\\texttt{Send\\_Basetype\\_Number} resp.
\\texttt{Receive\\_Basetype\\_Number} in the parameter file.
\\subsubsection*{Number 30}
\\begin{spacing}{1}
\\begin{verbatim}
    MPI_Type_contiguous(params.blocks * params.blocksize,
                        _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 31}
\\begin{spacing}{1}
\\begin{verbatim}
    MPI_Type_vector(params.blocks,params.blocksize,
                    params.vectorstride,
                    _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 32}
\\begin{spacing}{1}
\\begin{verbatim}
    MPI_Type_create_hvector(params.blocks,params.blocksize,
                            params.vectorstride*
                            _basetype_size[bt_number],
                            _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 33}
\\begin{spacing}{1}
\\begin{verbatim}
    blen[0] = params.blocksize;
    disp[0] = 0;
    k = 1;
    for (i=1; i<params.blocks; i++)
    {
      blen[i] = params.blocksize-k;
      disp[i] = disp[i-1]+blen[i-1]+1;
      k = 1-k;
    }
    blen[params.blocks-1] += params.blocks/2;
    MPI_Type_indexed(params.blocks,blen,disp,
                     _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 34}
\\begin{spacing}{1}
\\begin{verbatim}
    /* requires MPI 2 */
    for (i=0; i<params.blocks; i++)
    {
      disp[i] = params.blocksize*i+i;
    }
    MPI_Type_create_indexed_block(params.blocks,
                                  params.blocksize,disp,
                                  _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 35}
\\begin{spacing}{1}
\\begin{verbatim}
    blen[0] = params.blocksize;
    offset[0] = 0;
    k = 1;
    for (i=1; i<params.blocks; i++)
    {
      blen[i] = params.blocksize-k;
      offset[i] = offset[i-1]+blen[i-1]+1;
      k = 1-k;
    }
    for (i=0; i<params.blocks; i++)
    {
      offset[i] *= _basetype_size[bt_number];
    }
    blen[params.blocks-1] += params.blocks/2;

    MPI_Type_create_hindexed(params.blocks,blen,offset,
                             _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 36}
\\begin{spacing}{1}
\\begin{verbatim}
    blen[0] = params.blocksize;
    offset[0] = 0;
    type[0] = _basetype[bt_number];
    k = 1;
    for (i=1; i<params.blocks; i++)
    {
      blen[i] = params.blocksize-k;
      offset[i] = offset[i-1]+blen[i-1]+1;
      type[i] = _basetype[bt_number];
      k = 1-k;
    }
    for (i=0; i<params.blocks; i++)
    {
      offset[i] *= _basetype_size[bt_number];
    }
    blen[params.blocks-1] += params.blocks/2;

    MPI_Type_create_struct(params.blocks,blen,offset,type,d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 37}
\\begin{spacing}{1}
\\begin{verbatim}
    /* requires MPI 2, especially subarrays */
    for (i=0; i<params.blocks; i++)
    {
      blen[i] = SUBARRAYOFFSET+params.blocksize+SUBARRAYOFFSET;
      sublen[i] = params.blocksize;
      disp[i] = SUBARRAYOFFSET;
    }
    MPI_Type_create_subarray(params.blocks,blen,sublen,disp,
                             MPI_ORDER_C,_basetype[bt_number],
                             d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 50}
\\begin{spacing}{1}
\\begin{verbatim}
      MPI_Type_vector(len*params.blocks,params.blocksize,
                      params.vectorstride,
                      _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 51}
\\begin{spacing}{1}
\\begin{verbatim}
      MPI_Type_vector(len,params.blocks*params.blocksize,
                      params.blocks*params.vectorstride,
                      _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 52}
\\begin{spacing}{1}
\\begin{verbatim}
      MPI_Type_vector(params.blocks,len*params.blocksize,
                      len*params.vectorstride,
                      _basetype[bt_number],d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 53}
\\begin{spacing}{1}
\\begin{verbatim}
      blen[0] = params.blocksize;
      offset[0] = 0;
      type[0] = _basetype[bt_number];
      k = 1;
      for (i=1; i<len*params.blocks; i++)
      {
        blen[i] = params.blocksize-k;
        offset[i] = offset[i-1]+blen[i-1]+1;
        type[i] = _basetype[bt_number];
        k = 1-k;
      }
      for (i=0; i<len*params.blocks; i++)
      {
        offset[i] *= _basetype_size[bt_number];
      }
      blen[i-1] += i/2;
      MPI_Type_create_struct(len*params.blocks,blen,offset,
                             type,d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 54}
\\begin{spacing}{1}
\\begin{verbatim}
      blen[0] = params.blocksize;
      offset[0] = 0;
      type[0] = _basetype[bt_number];
      k = 1;
      for (i=1; i<len; i++)
      {
        blen[i] = params.blocks*params.blocksize-k;
        offset[i] = offset[i-1]+blen[i-1]+1;
        type[i] = _basetype[bt_number];
        k = 1-k;
      }
      for (i=0; i<len; i++)
      {
        offset[i] *= _basetype_size[bt_number];
      }
      blen[i-1] += i/2;
      MPI_Type_create_struct(len,blen,offset,type,d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 55}
\\begin{spacing}{1}
\\begin{verbatim}
      blen[0] = params.blocksize;
      offset[0] = 0;
      type[0] = _basetype[bt_number];
      k = 1;
      for (i=1; i<params.blocks; i++)
      {
        blen[i] = len*params.blocksize-k;
        offset[i] = offset[i-1]+blen[i-1]+1;
        type[i] = _basetype[bt_number];
        k = 1-k;
      }
      for (i=0; i<params.blocks; i++)
      {
        offset[i] *= _basetype_size[bt_number];
      }
      blen[i-1] += i/2;
      MPI_Type_create_struct(params.blocks,blen,offset,type,d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 56}
\\begin{spacing}{1}
\\begin{verbatim}
      MPI_Type_vector(params.blocksize,len,
                      len+params.vectorstride,
                      _basetype[bt_number],&bvector1);
      MPI_Type_commit(&bvector1); /* superfluous(?),
                                     but why not */
      MPI_Type_vector(params.blocks,1,2,bvector1,d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 57}
\\begin{spacing}{1}
\\begin{verbatim}
      MPI_Type_vector(len,params.blocksize,params.vectorstride,
                      _basetype[bt_number],&bvector2);
      MPI_Type_commit(&bvector2); /* superfluous(?),
                                     but why not */
      MPI_Type_vector(params.blocks,1,2,bvector2,d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 58}
\\begin{spacing}{1}
\\begin{verbatim}
      MPI_Type_vector(params.blocks,params.blocksize,
                      params.vectorstride,_basetype[bt_number],
                      &bvector3);
      MPI_Type_commit(&bvector3); /* superfluous(?),
                                     but why not */
      MPI_Type_vector(len,1,2,bvector3,d);
\\end{verbatim}
\\end{spacing}
\\subsubsection*{Number 70}
\\begin{spacing}{1}
\\begin{verbatim}
    /* Special types (assume here that BLOCKSIZE>=3) */

    MPI_Type_vector(params.blocks,1,params.blocksize,
                    _basetype[bt_number],&comp1);
    MPI_Type_commit(&comp1);
    for (i=0; i<params.blocks; i++) {
      blen[i] = 1;
      disp[i] = i*params.blocksize;
    }
    MPI_Type_indexed(params.blocks,blen,disp,
                     _basetype[bt_number],&comp2);
    MPI_Type_commit(&comp2);
    for (i=0; i<params.blocks; i++) {
      blen[i] = params.blocksize-2;
      offset[i] = i*params.blocksize*_basetype_size[bt_number];
      type[i] = _basetype[bt_number];
    }
    #if MPI_2
    MPI_Type_create_struct(3,blen,offset,type,&comp3);
    #else
    MPI_Type_struct(3,blen,offset,type,&comp3);
    #endif
    MPI_Type_commit(&comp3);
    blen[0] = 1; offset[0] = 0; type[0] = comp1;
    blen[1] = 1; offset[1] = _basetype_size[bt_number];
    type[1] = comp2;
    blen[2] = 1; offset[2] = 2*_basetype_size[bt_number];
    type[2] = comp3;
    #ifdef MPI_2
    MPI_Type_create_struct(3,blen,offset,type,d);
    #else
    MPI_Type_struct(3,blen,offset,type,d);
    #endif
\\end{verbatim}
\\end{spacing}

\\end{spacing}
EOT
}
#### eo print_basetypes_info

####
# print_output_file_trailer
#
# Prints the end of the resulting LaTeX file.
#
# PARAMS:
#   none
# RETURNS:
#   nothing
####
sub print_output_file_trailer
{
    print TEX_FILE <<'EOT';
\end{spacing}
\end{document}
EOT
}

######## headers for each pattern ########

####
# print_p2p_header, print_mw_header, print_col_header, print_simple_header
#
# These functions print the headers for the pattern-specific sections.
#
# PARAMS:
#   none
# RETURNS:
#   nothing
####
sub print_p2p_header
{
    print TEX_FILE <<'EOT'
\section{Point-to-point-pattern}
\label{p2p_dummy}
The point-to-point-pattern calls the {\tt routine\_to\_be\_measured}
to communicate with the farest node and the nearest node. These calls
are varied over message length. Every parameter set is called
{\tt repetitions} times and the average value is stored. Since we are
programming a message-passing MIMD system, we have distinct code for
the server (measurement) and the client (just answering).


As result the sum of the incoming bandwidth and outgoing bandwidth is given.
can be described in pseudocode as follows:
\begin{spacing}{1}
\begin{verbatim}
/* Server-node */

max_node := node with maximum latency;

for message-length:=0 to maximum message-length step step-width do

  for i := 1 to N do
    start_time := MPI_Wtime();
    routine_to_be_measured (max_node, message_length);
    end_time :=MPI_Wtime();
    time[i] := end_time - start_time;

  time := extract (time[1]..time[N]);   


/* Client code */

actions to answer the max/min_node determination;

if (I am the max_node)
  for message-length:=0 to maximum message-length step step-width do
    for i := 1 to repetitions do
      client answer for the routine_to_be_measured (message_length);

\end{verbatim}
\end{spacing}
EOT
}
#### eo print_p2p_header ####

sub print_mw_header
{
    print TEX_FILE <<'EOT'
\section{Master-worker-pattern}
\label{mw_dummy}
The Master-worker-pattern correspondents to the typical master-worker-scheme:
a master dispatches pieces of work to several workers. Each worker
does his piece of work. When finished he sends his result to the master and
requests for a new piece of work (and so on). When all work is done, the
master sends an stop-signal to the workers.\\
This scheme is important in practise, since it automatically balances
load. In pseudo-code the Master-worker-scheme looks like:
\begin{spacing}{1}
\begin{verbatim}
/* master-code */

for each worker
  set ready to receive;   /* e.g. MPI_Irecv */

chunk :=0;
start_time := MPI_Wtime;

while chunk < all_chunks
  dispatch (chunk, msglen);
  chunks := chunks + 1;

end_time := MPI_Wtime;

for each worker
  send stop signal;


/* worker-code */

forever
  send ready signal to master;

  receive signal (msglen);

  if signal == stop signal
    exit;

  do work;       /* corresponding to the received signal */
  send result;

endforever\end{verbatim}
\end{spacing}
Every abstract communication ``code'' in the scheme above can be filled
with concrete MPI\_Code. We measure the time consumed by {\tt dispatch work}.
This code sequence does for example this:
\begin{spacing}{1}
\begin{verbatim}
/* dispatch work: */
wait for a worker;
receive work from worker;
send actual piece of work to worker;
set ready to receive next piece of work from worker;
actual piece of work := next piece of work;
\end{verbatim}
\end{spacing}

EOT
}
#### eo print_mw_header ####

sub print_col_header
{
    print TEX_FILE <<'EOT'
\section{Collective-pattern}
\label{col_dummy}
We want to use the following pattern to measure collective operations:
\begin{spacing}{1}
\begin{verbatim}
/* server-code */

MPI_Barrier;
for i := 1 to N do
  start_time := MPI_Wtime();
  routine_to_be_measured;
  MPI_Barrier;
  end_time := MPI_Wtime();
  time[i] := end_time - start_time;

time := extract(time[1]..time[N]); 

/* client code */

MPI_Barrier;
for i := 1 to N do
  client_routine;   /* as answer for routine_to_be_measured */
  MPI_Barrier();
\end{verbatim}
\end{spacing}

EOT
}
#### eo print_col_header ####

sub print_simple_header
{
    print TEX_FILE <<'EOT'
\section{Simple-pattern}
\label{simple_dummy}
Some routines seem to be so simple, that they are measured in a very simple
``pattern''.
\begin{spacing}{1}
\begin{verbatim}

for i := 1 to N do
  start_time := MPI_Wtime();
  routine_to_be_measured;
  end_time := MPI_Wtime();
  time[i] := end_time - start_time;

time := extract(time[1]..time[N]);
\end{verbatim}
\end{spacing}

EOT
}
#### eo print_simple_header ####

######## eo headers for each pattern ########

####
# rsplit
#
# Splits the `skampi.out' file into `.ski' files one of which represents
# one suite of measurements.
# Existing additional header parts with information of a SKaMPI run are
# processed so that comparisons between other machines are possible.
# The filenames of these additional `.skic' files must end (before the `.skic')
# with a serial number per run.
# Reliable results are guaranteed only in combination with the Java applet
# which implements the interface to the SKaMPI results database.
# NOTE: Gnuplot crashes (i. e. creates BIIIIG .eps files) when confronted with
# too long lines. So, in addition to the `.ski' and `.skic' files `.cut' files
# are produced which have lines short enough for gnuplot.
#
# PARAMS:
#   String - filename of `skampi.out' file
# RETURNS:
#   nothing
####
sub rsplit
{
    my $data_file = shift @_;
    my $open_flag = 0;
    my $run       = 0;
    my $next_run  = 0;
    my $hash_ref;
    my $key;
    my $value;
    my @scales    = ();  # used to cut logarithmic 0 args from data

    open(IN_FILE, "<$data_file")
      || quit("Cannot open SKaMPI data file `$data_file': $!", __LINE__);

    my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
    while (<IN_FILE>)
    {
        if (/^\/\*\@next\*\//)
        {
            my %additional_machine_info = parse_run_info(*IN_FILE);

            foreach (keys(%additional_machine_info))
            {
                $additional_machine_info{$_} = '?'
                  if ($additional_machine_info{$_} eq '');

                $additional_machine_info{$_} =
                  convert2tex($additional_machine_info{$_});
            }

            push(@comp_info_hash_list, \%additional_machine_info);
        }
        if (/\/\*\@in(\D.*?)\*\//)
        {
            $file_name = $1;

            close(OUT_FILE)     if ($open_flag);
            close(CUT_OUT_FILE) if ($open_flag);

            # is this file relevant?
            if ($file_name =~ /\.ski$/)
            {
                unless (($report_all_suites &&
                         !grep($file_name eq $_, @excluded_suites)) ||
                        grep($file_name eq $_,
                             (@reported_suites, @compared_suites)))
                {
                    $open_flag = 0;

                    next;
                }
            }
            elsif (!grep($file_name eq $_, @compared_suites))
            {
                $open_flag = 0;

                next;
            }

            grep($file_name eq $_, @files_to_delete)
              && quit("$data_file (line $.): Existing file `$file_name' ".
                      "would be overwritten", __LINE__);
            grep("$file_name.cut" eq $_, @files_to_delete)
              && quit("$data_file (line $.): Existing file `$file_name.cut' ".
                      "would be overwritten", __LINE__);
            open(OUT_FILE, ">$file_name")
              || quit("$data_file (line $.): Cannot open `$file_name'",
                      __LINE__);
            open(CUT_OUT_FILE, ">$file_name.cut")
              || quit("$data_file (line $.): Cannot open `$file_name.cut'",
                      __LINE__);
            push(@files_to_delete, $file_name, "$file_name.cut");
            myprint("$data_file: Opening file `$file_name'\n".
                    "$data_file: Opening file `$file_name.cut'\n");
            $open_flag = 1;
            @scales    = ();
            next;
        }

        if (/\/\*\@no\*\//)
        {
            if ($open_flag)
            {
                close(OUT_FILE)
                  || quit("$data_file (line $.): Cannot close `$file_name': ".
                          $!, __LINE__);
                close(CUT_OUT_FILE)
                  || quit("$data_file (line $.): Cannot close ".
                          "`$file_name.cut': $!", __LINE__);
                myprint("$data_file: Closing file `$file_name'\n".
                        "$data_file: Closing file `$file_name.cut'\n");
                $open_flag = 0;
            }
            next;
        }

        if (/\/\*\@ap.*\*\//)
        {
            quit("$data_file (line $.): Cannot append to already opened ".
                 "file `$file_name'", __LINE__)
              if $open_flag;

            $file_name = $1 if (/\/\*\@ap(\D.*?)\*\//);

            open(OUT_FILE,">>$file_name")
              || quit("$data_file (line $.): Cannot append to `$file_name': ".
                      $!, __LINE__);
            open(CUT_OUT_FILE,">>$file_name.cut")
              || quit("$data_file (line $.): Cannot append to ".
                      "`$file_name.cut': $!", __LINE__);
            myprint("$data_file: Appending to file `$file_name'\n".
                    "$data_file: Appending to file `$file_name.cut'\n");
            $open_flag = 1;
            next;
        }

        # initialize list of scales if it's an appropriate line
        @scales = parse_scale_line($_);

        if (my @numbers = parse_measurement_line($_))
        {
            my $line    = join(' ', @numbers)."\n";
            my $cut_line;
            my $count   = 0;

            # trim line to argument and result values (for gnuplot)
            $cut_line =
              join(' ', @numbers[0..(5 + (@scales ? @scales : 1))])."\n";

            # put line in comment if some logarithmic argument is 0;
            # this is due to a problem of gnuplot
            foreach (@scales)
            {
                if (/logarithmic/ && $numbers[$count]==0)
                {
                    $cut_line = "##".$cut_line;
                    last;
                }
                $count++;
            }

            print(CUT_OUT_FILE $cut_line) if ($open_flag);

            $_ = $line;
        }

        print(OUT_FILE $_) if ($open_flag);
    }
    $_ = $saved_arg;

    close(OUT_FILE)     if ($open_flag);
    close(CUT_OUT_FILE) if ($open_flag);
    close(IN_FILE);
}
#### eo rsplit ####

####
# read_header
#
# Parses the header part of one suite file and returns a list of the values
# describing this suite.
#
# PARAMS:
#   String - filename of `.ski' file
# RETURNS:
#   List   - information about suite:
#            [0]  - filename (`p2p_MPI_Send-MPI_Recv.ski' etc.)
#            [1]  - measurement name (`MPI_Send-MPI_Recv' etc.)
#            [2]  - pattern (`Point-to-Point' etc.)
#            [3]  - LIST REF.: -> variation params (`message length' etc.)
#            [4]  - LIST REF.: -> units of params (`Bytes' etc.)
#            [5]  - LIST REF.: -> formats of params (`%d' etc.)
#            [6]  - LIST REF.: -> scales of params (`logarithmic' etc.)
#            [7]  - LIST REF.: -> refinements of params
#                                 (`automatic x wide adaption' etc.;
#                                  each can be undef)
#            [8]  - LIST REF.: -> range starts (integer)
#            [9]  - LIST REF.: -> range ends (integer)
#            [10] - LIST REF.: -> range stepwidths (float)
#            [11] - default values
#                   (`default values: 32 nodes, max. / act. time ...' etc.)
#            [12] - standard error (float)
#            [13] - cut quantile (float)
#            pattern "Simple" => [3] - [11] empty lists or string
####
sub read_header  # of a single suite of measurements
{
    my $file_name = shift @_;
    my @ms;
    my %suite_hash;
    my @suite_info;

    open(SKI_FILE, "<$file_name")
      || quit("Cannot open `$file_name': $!", __LINE__);

    %suite_hash = parse_suite_info(*SKI_FILE);
    @suite_info = create_suite_info_list(\%suite_hash);

    close(SKI_FILE);

    @ms = ($file_name, @suite_info[0..12]);

    # delete underscores in variation parameters
    foreach (0..$#{$ms[3]})
    {
        $ {$ms[3]}[$_] =~ tr/_/ /;
    }

    # adjust default values
    if (defined($ms[11]))
    {
        $ms[11] = "default values: $ms[11]";
    }
    else
    {
        $ms[11] = '';
    }

    return (@ms);
}
#### eo read_header ####

####
# header2tex_module
#
# Creates a LaTeX file - which is included in the main LaTeX file - out of
# the information of one suite.
#
# PARAMS:
#   List - [0]-[13] - s. return value of read_header
#          [14]     - reference to list of filenames for `.eps' files to be
#                     included
#          [15]     - reference to list of filenames for `.eps' files for
#                     node times to be included
# RETURNS:
#   $ - 0: .tmd file has not been produced
#       1: .tmd file has been produced
####
sub header2tex_module
{
    my ($ski_file,
        $name,
        $pattern_str,
        $var_strs_ref,
        $unit_strs_ref,
        $format_strs_ref,
        $scale_strs_ref,
        $refine_strs_ref,
        $starts_ref,
        $ends_ref,
        $steps_ref,
        $def_values,
        $standard_error,
        $cut_quantile,
        $result_files_ref,
        $node_files_ref)   = @_;
    my $file_name = "$pattern_str_hash{$pattern_str}_$name.tmd";
    my @data_list;

    $ {$pattern_existing{$pattern_str}} = 1;

    if (-f $file_name && !$overwrite)
    {
        myprint ("Using existing TeX module file `$file_name'\n");

        return 1;
    }

    open(TMD_FILE,">$file_name")
      || quit("Cannot open `$file_name': $!", __LINE__);
    push(@files_to_delete, $file_name);

    my $quoted_name = convert2tex($name);

    # read list of results
    @data_list = read_data($ski_file);

    # were there any data lines at all?
    unless (@data_list)
    {
        mywarn("  $name: no data lines\n");

        return 0;
    }

    # compute bandwidths according to measured function
    my $bdw_str1 = "";
    my $bdw_str2 = "";

    if (grep("message length" eq $_, @$var_strs_ref))
    {
        my ($bdw_func_ref,
            $param)        = determine_bdw_func ($ski_file,
                                                 $name,
                                                 $pattern_str,
                                                 $var_strs_ref,
                                                 $def_values);

        if (defined $bdw_func_ref)
        {
            $bdw_str1 = comp_bdw($bdw_func_ref,
                                 $data_list[0],
                                 $pattern_str,
                                 $var_strs_ref,
                                 $param);
            $bdw_str2 = comp_bdw($bdw_func_ref,
                                 $data_list[-1],
                                 $pattern_str,
                                 $var_strs_ref,
                                 $param);
        }
    }  # message length exists

    if ($pattern_str ne "Simple")
    {
        print TMD_FILE <<"EOT";
\\newpage
\\subsection{$quoted_name}
  \\hspace{1em}
EOT
        # dri == dorep information
        if (open (DRI_FILE,"<$dri_directory$name.dri"))
        {
            my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
            while (<DRI_FILE>)
            {
                print TMD_FILE $_;
            }
            $_ = $saved_arg;
            close (DRI_FILE);
        }
        else
        {
            # Don't print a warning if there's not a single dri file around.
            # This has been pointed out earlier in the program.
            mywarn ("Cannot open dri-file: $dri_directory$name.dri\n")
              unless ($dri_files_missing);
        }

        # if reference to file list undefined there are too many data lines
        # => don't even create a table
        if (defined($result_files_ref))
        {
            my @result_files = @$result_files_ref;

            if (@result_files)
            {
                foreach (@result_files)
                {
                    print(TMD_FILE <<"EOT");
  \\begin{center}
    \\epsfig{file=$_.eps}
  \\end{center}
EOT
                }
            }
            else  # no eps files => table
            {
                my $tex_columns = "|r" x (@$var_strs_ref + 3)."|";
                my $arguments   = join(" & ",
                                       map(convert2tex($_), @$var_strs_ref));

                print(TMD_FILE <<"EOT");
%\\begin{figure}[Htb]
\\begin{center}
%\\newline
\\begin{tabular}{$tex_columns}
\\hline
$arguments & Result [\$\\mu\$s] & Std. error [\$\\mu\$s] & Count\\\\
\\hline
EOT
                foreach (@data_list)
                {
                    my @entries = (@{$$_[0]}, @$_[1..3]);
                    my $table_row = join(" & ", @entries);

                    print(TMD_FILE "$table_row\\\\\n");
                }

                print(TMD_FILE <<EOT);
\\hline
\\end{tabular}
%\\caption{Measurement $quoted_name varied over @$var_strs_ref.}
%\\label{$name}
\\end{center}
%\\end{figure}
EOT
            }  # table instead of eps file
        }  # eps files or table

        @node_files = @$node_files_ref;
        if (@node_files)
        {
            my $eps_count = 0;

            # allow page break between wide eps files by centering each single
            # image
            print(TMD_FILE "  \\begin{center}\n") unless ($wide_nt);

            foreach (@node_files)
            {
                # end centering after every second small image
                # to allow page break there
                if (!$wide_nt && $eps_count && !($eps_count % 2))
                {
                    print(TMD_FILE "  \\end{center}\n");
                    print(TMD_FILE "  \\begin{center}\n");
                }

                print(TMD_FILE "  \\begin{center}\n") if ($wide_nt);

                print TMD_FILE <<"EOT";
    \\epsfig{file=$_.eps}
EOT

                print(TMD_FILE "  \\end{center}\n") if ($wide_nt);

                $eps_count++;
            }

            print(TMD_FILE "  \\end{center}\n") unless ($wide_nt);
        }  # eps files for node times

        print_suite_info(TMD_FILE,
                         $ski_file,
                         $name,
                         $pattern_str,
                         $var_strs_ref,
                         $unit_strs_ref,
                         $format_strs_ref,
                         $scale_strs_ref,
                         $refine_strs_ref,
                         $starts_ref,
                         $ends_ref,
                         $steps_ref,
                         $def_values,
                         $standard_error,
                         $cut_quantile);
        if ($bdw_str1 ne "" || $bdw_str2 ne "")
        {
            print(TMD_FILE <<"EOT");
\\begin{itemize}
  $bdw_str1
  $bdw_str2
\\end{itemize}
\\newpage
EOT
        }
        else
        {
            print(TMD_FILE "\\newpage\n");
        }
    }
    else  # simple
    {
        print(TMD_FILE <<EOT);
\\subsection{$quoted_name}
EOT
        # dri == dorep information
        if (open (DRI_FILE,"<$dri_directory$name.dri"))
        {
            my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
            while (<DRI_FILE>)
            {
                print(TMD_FILE $_);
            }
            $_ = $saved_arg;
            close(DRI_FILE);
        }
        else
        {
            # Don't print a warning if there's not a single dri file around.
            # This has been pointed out earlier in the program.
            mywarn ("Cannot open dri-file: $dri_directory$name.dri\n")
              unless ($dri_files_missing);
        }

        print(TMD_FILE <<EOT);
\\begin{itemize}
\\setlength{\\itemsep}{0mm}
\\setlength{\\parsep}{0mm}
  \\item Pattern: $pattern_str
  \\item Max. allowed standard error is $standard_error \\%
\\end{itemize}
%\\begin{figure}[Htb]
\\begin{center}
%\\newline
\\begin{tabular}{|r|r|r|}
\\hline
Result [\$\\mu\$s] & Standard error [\$\\mu\$s] & Count\\\\
\\hline
EOT
        foreach (@data_list)
        {
            my ($result, $std_error, $count) = @$_[1..3];

            print(TMD_FILE "$result & $std_error & $count\\\\\n");
        }
        print(TMD_FILE <<EOT);
\\hline
\\end{tabular}
%\\caption{Measurement $quoted_name.}
%\\label{$name}
\\end{center}
%\\end{figure}
EOT
    }
    close(TMD_FILE);

    return 1;
}
#### eo header2tex_module ####

####
# determine_bdw_func
#
# Determines the function with which the bandwidth of a measurement has to
# be calculated.
#
# PARAMS:
#   List - [0] `.ski' filename
#          [1] suite name
#          [2] pattern
#          [3] ref. to list of variation parameters
#          [4] default values
# RETURNS:
#   List - [0] ref. to function; [1] optional parameter
####
sub determine_bdw_func
{
    my ($ski_file,
        $name,
        $pattern_str,
        $var_strs_ref,
        $def_values)   = @_;
    my $bdw_func_ref = undef;
    my $param        = undef;

    if ($pattern_str eq "Point-to-Point")
    {
        $bdw_func_ref = \&comp_bdw_p2p;
        $param        = undef;  # not used with Point-to-Point
    }
    elsif ($pattern_str eq "Master-Worker")
    {
        $bdw_func_ref = \&comp_bdw_mw;

        if (grep("chunks" eq $_, @$var_strs_ref))
        {
            $param = undef;  # must be computed per line
        }
        elsif ($def_values =~ /chunks (\d+)/)
        {
            $param = $1;
        }
        else
        {
            mywarn ("Number of chunks undefined in `$ski_file' ".
                    "- assuming 1 chunk\n");
            $param = 1;
        }
    }  # master-worker
    elsif ($pattern_str eq "Collective")
    {
        # determine the function to compute the factor for
        # aggregate bandwidth
        # see Hwang, Scalable Parallel Computing, p. 591
        $bdw_func_ref = \&comp_bdw_col_sng;  # default

        # examine name for reg. expr. in hash
        foreach (keys %band_width_func)
        {
            if ($name =~ /$_/)
            {
                $bdw_func_ref = $band_width_func{$_};
                last;
            }
        }

        if (grep("number of nodes" eq $_, @$var_strs_ref))
        {
            $param = undef;  # must be computed per line
        }
        elsif ($def_values =~ /(\d+) nodes/)
        {
            $param = $1;
        }
        else
        {
            mywarn ("Number of nodes undefined in suite `$ski_file' ".
                    "- assuming 1 node\n");
            $param = 1;
        }
    }  # collective
    else  # function undefined
    {
        mywarn ("Unknown pattern `$pattern_str'\n");
    }

    return ($bdw_func_ref, $param);
}
#### eo determine_bdw_func

####
# print_suite_info
#
# Prints information about a suite of measurements such as pattern, variation
# parameters, ranges etc.
#
# PARAMS:
#   Filehandle - handle to file to which info shall be printed
#   List       - s. return value of read_header
# RETURNS:
#   nothing
####
sub print_suite_info
{
    my ($filehandle,
        $ski_file,
        $name,
        $pattern_str,
        $var_strs_ref,
        $unit_strs_ref,
        $format_strs_ref,
        $scale_strs_ref,
        $refine_strs_ref,
        $starts_ref,
        $ends_ref,
        $steps_ref,
        $def_values,
        $standard_error,
        $cut_quantile)   = @_;
    map s/ /_/g, (my @var_strs_ = @$var_strs_ref);
    my @unit_strs   = @$unit_strs_ref;
    my $units       = '';  # description of the parameters' units
    my $var_str     = join(", ", map(convert2tex($_), @$var_strs_ref));
    my $tex_columns = "|l". "|c" x (@$var_strs_ref)."|";
    my $var_row     = join(" & ", map(convert2tex($_), @$var_strs_ref));
    my @scales      = @$scale_strs_ref;
    my @refines     = @$refine_strs_ref;
    my $scale_row   = "";
    my $refine_row  = "";
    my $range_row   = "";
    my $step_row    = join " & ", @$steps_ref;
    my $counter     = 0;

    # create correct unit strings
    foreach (@unit_strs)
    {
        if (defined($unit_hash{$var_strs_[$counter]}) &&
            $_ eq $ {$unit_hash{$var_strs_[$counter]}}[0])
        {
            $_ = $ {$unit_hash{$var_strs_[$counter]}}[1];
        }
        $counter++;
    }

    $scale_row = join(" & ", map(convert2tex($_), @scales));
    if (grep(defined(), @refines))
    {
        map({$_ = "---" if (!defined)} @refines);
        $refine_row  = "Param. refinement  & ";
        $refine_row .= join(" & ", map(convert2tex($_), @refines));
    }

    $counter = 0;
    foreach (@$starts_ref)
    {
        $range_row .= " & $_ -- $$ends_ref[$counter] ".
          sprintf("units %s(s. below)",
                  @$starts_ref == 1 ? '' : '('.($counter + 1).') ');
        $counter++;
    }

    # create description of units
    foreach (@unit_strs)
    {
        $units .= '\\item ' if (@unit_strs > 1);  # multi-dimensional
        $units .= convert2tex($_)."\n";
    }
    $units = "\\begin{enumerate}\n$units\\end{enumerate}\n"
      if (@unit_strs > 1);

    $def_values = convert2tex($def_values);

    print $filehandle <<"EOT";
\\begin{itemize}
\\setlength{\\itemsep}{0mm}
\\setlength{\\parsep}{0mm}
  \\item Pattern: $pattern_str varied over $var_str.
  \\item $def_values
  \\item Max. allowed standard error is $standard_error \\%.
\\end{itemize}
\\begin{center}
  \\begin{tabular}{$tex_columns}
    \\hline
                       & $var_row\\\\
    \\hline
    X axis scale       & $scale_row\\\\
    $refine_row\\\\
    Argument range    $range_row\\\\
    Argument stepwidth & $step_row\\\\
    \\hline
  \\end{tabular}
\\end{center}
\\begin{itemize}
  \\item Units:
    $units
\\end{itemize}
EOT
}
#### eo print_suite_info ####

####
# header2gpl_files
#
# Produces some (currently not more than 1) gnuplot files which gnuplot uses
# to produce `.eps' files which show the suite's data.
#
# PARAMS:
#   List - s. return value of read_header
# RETURNS:
#   Reference - ref. to list of filenames (without extension) to be gnuplotted
#               (list is empty if not enough data lines)
#   undef     - if not even a table shall be created in the LaTeX file
####
sub header2gpl_files
{
    my ($ski_file,
        $name,
        $pattern_str,
        $var_strs_ref,
        $unit_strs_ref,
        $format_strs_ref,
        $scale_strs_ref,
        $refine_strs_ref,
        $starts_ref,
        $ends_ref,
        $steps_ref,
        $def_values,
        $standard_error,
        $cut_quantile)   = @_;
    # filename for gnuplot file
    my $file_name     = "$pattern_str_hash{$pattern_str}_$name";
    my @to_be_created = ();  # files to be created (returned)

    # simple suites consist of no more than one line
    return [] if ($pattern_str eq "Simple");

    my @data_list = read_data($ski_file);

    if (@data_list < $minimum_data)
    {
        return [];  # not enough data lines for suite `.eps' file
    }

    if (@$var_strs_ref > 2)
    {
        myprint("$file_name: no graphical output supported for ".
                scalar(@$var_strs_ref)."-dimensional data\n");

        return undef;  # don't even create a table
    }

    if (-f "$file_name.gpl" && !$overwrite)
    {
        myprint("Using existing gnuplot file `$file_name.gpl'\n");

        return [$file_name];
    }

    open(GPL_FILE,">$file_name.gpl")
      || quit("Cannot open `$file_name.gpl': $!", __LINE__);
    push(@files_to_delete, "$file_name.gpl");

    my $output3d = "";
    $output3d = "set parametric\nset dgrid3d" if (@$var_strs_ref == 2);

    my $label_str = "set xlabel \"$$var_strs_ref[0]\"\n".
      "set ylabel ";
    if (@$var_strs_ref == 2)
    {
        $label_str .= "\"$$var_strs_ref[1]\"\n".
          "set zlabel ";
    }
    $label_str .= '"time (microsec.)"';

    my $x_log_scale = "";
    my $y_log_scale = "";
    my $z_log_scale = "";
    if ($$scale_strs_ref[0] =~ /logarithmic/)
    {
        $x_log_scale = "set logscale x 10";
        $y_log_scale = "set logscale y 2";
        $z_log_scale = "set logscale z 2";
    }
    if (@$var_strs_ref == 2)
    {
        if ($$scale_strs_ref[1] =~ /logarithmic/)
        {
            $y_log_scale = "set logscale y 10";
            $z_log_scale = "set logscale z 2";
        }
        else
        {
            $y_log_scale = "";
        }
    }

    my $plot_command = "";
    if (@$var_strs_ref == 1)
    {
        $plot_command = "plot [$$starts_ref[0]:$$ends_ref[0]] ".
          "\"$ski_file.cut\" title \"$name\" with linesp, ".
          "\"$ski_file.cut\" using 1:2:3 title \"\" with errorbars\n";
    }
    else
    {
        $plot_command = "splot [$$starts_ref[0]:$$ends_ref[0]] ".
          "[$$starts_ref[1]:$$ends_ref[1]] ".
          "\"$ski_file.cut\" title \"$name\" with lines\n";
    }

    print GPL_FILE <<EOT;
#gnuplot command-file for the measurement $name
set nologscale
$output3d
set xtics
set ytics
set ztics
set term postscript eps 18
$label_str
$x_log_scale
$y_log_scale
$z_log_scale
set output  # to be redirected to $file_name.eps
$plot_command
EOT
    close (GPL_FILE);

    push (@to_be_created, $file_name);

    return \@to_be_created;
}
#### eo header2gpl_files ####

####
# header2node_gpl_files
#
# Produces some gnuplot files which gnuplot uses to produce `.eps' files
# which show some of the suite's node times in a barchart.
#
# PARAMS:
#   List - s. return value of read_header
# RETURNS:
#   Reference - ref. to list of filenames (without extension) to be gnuplotted
####
sub header2node_gpl_files
{
    my ($ski_file,
        $name,
        $pattern_str,
        $var_strs_ref,
        $unit_strs_ref,
        $format_strs_ref,
        $scale_strs_ref,
        $refine_strs_ref,
        $starts_ref,
        $ends_ref,
        $steps_ref,
        $def_values,
        $standard_error,
        $cut_quantile)   = @_;
    my $node_file     = "";           # filename of gnuplot file for node times
    my $file_count    = 0;            # counter for node time files
    my $data_count    = 0;            # counter for data lines
    my @to_be_created = ();           # files to be created (returned)

    # simple suites consist of no more than one line without node times
    return [] if ($pattern_str eq "Simple");

    my @data_list = read_data($ski_file);

    # examine if node times should be shown in an `.eps' file
    foreach (@data_list)
    {
        my @args          = @{$$_[0]};
        my @sorted_data   = @{$$_[7]};  # node times are already sorted
        my $title_var_str = "";

        my $arg_count = 0;  # number of arguments which fit for barchart
        foreach (@$var_strs_ref)
        {
            my $var_str = $_;
            (my $var_str_ = $_) =~ s/ /_/g;

            if (!grep("none" eq $_, @{$barchart_values{$var_str_}}))
            {
                foreach (@{$barchart_values{$var_str_}})
                {
                    if (($_ eq "min" && $data_count == 0) ||
                        ($_ eq "max" && $data_count == $#data_list) ||
                        ($_ ne "min" && $_ ne "max" &&
                         $_ == $args[$arg_count]))
                    {
                        $title_var_str .= ", " if ($title_var_str ne "");
                        $title_var_str .= "$var_str $args[$arg_count]";
                        $arg_count++;
                        last;
                    }
                }
            }
        }

        # all arguments designated for barchart output?
        # test for maximum node time > 0 because of some bug in SKaMPI
        if ($arg_count==@$var_strs_ref && @sorted_data>0 && $sorted_data[0]>0)
        {
            my $image_size = 'set size     .5, .5';

            $node_file = sprintf ("$pattern_str_hash{$pattern_str}_$name".
                                  ".nt%.3d", $file_count);
            open(NODE_FILE, ">$node_file")
              || quit("Cannot open `$node_file': $!", __LINE__);
            push(@files_to_delete, $node_file);
            push(@to_be_created,   $node_file);

            # write sorted node times to file
            my $node_count = 1;
            foreach (@sorted_data)
            {
                print(NODE_FILE "$node_count $_\n");
                $node_count++;
            }
            $node_count -= .5;  # for range end in gnuplot file

            close(NODE_FILE);

            if (-f "$node_file.gpl" && !$overwrite)
            {
                myprint("Using existing gnuplot file `$node_file.gpl'\n");
                $file_count++;
                $data_count++;

                next;
            }

            # give gnuplot some y-range if all node times are equal
            my $yrange = '';
            if ($sorted_data[0] == $sorted_data[-1])
            {
                $yrange = '['.($sorted_data[0]-.5).':'.($sorted_data[0]+.5).
                  ']';
            }

            # produce wider image if lots of node times will be plotted
            if (@sorted_data >= $min_node_times)
            {
                $image_size = 'set size     1, 1';

                $wide_nt = 1;  # for header2tex_module to allow page break
            }

            open(GPL_FILE, ">$node_file.gpl")
              || quit("Cannot open `$node_file.gpl': $!", __LINE__);
            push(@files_to_delete, "$node_file.gpl");

            print(GPL_FILE <<"EOT");
# gnuplot file for node times of measurement $name
set term     postscript eps 10
$image_size
set nogrid  #   ytics
set xtics    1, 1
set boxwidth .5
set xlabel   "node number (sorted)"
set ylabel   "time (microsec.)"
set title    "node times for $title_var_str"
set output  # to be redirected to $node_file.eps
plot [.5:$node_count] $yrange "$node_file" using 1:2 title "" with boxes
EOT
            close(GPL_FILE);

            $file_count++;
        }  # create barchart

        $data_count++;
    }  # loop over data lines

    return \@to_be_created;
}
#### eo header2node_gpl_files ####

######## bandwidth computation functions ########

####
# comp_bdw
#
# Produces a string which shows information about the bandwidth of a
# measurement as function of the message length. The units for the
# message length and the bandwidth are shown in units(/s) or in 1024 units(/s)
# depending on the values.
#
# PARAMS:
#   Reference - ref. to the function which calculates the bandwidth
#   Reference - ref. to list of numbers of a measurement (s. read_data)
#   String    - pattern name
#   Reference - ref. to list of variation parameters
#   Number    - optional parameter for the bandwidth function
# RETURNS:
#   created string (must be used within an itemize environment)
####
sub comp_bdw
{
    my ($bdw_func_ref,
        $numbers_ref,
        $pattern,
        $var_strs_ref,
        $param)        = @_;
    my @numbers    = @$numbers_ref;
    my $length     = 0;
    my $length_str = 'units';
    my $time       = 0;
    my $counter    = 0;
    my %var_hash   = ();
    my $bandwidth  = 0;
    my $bdw_str    = 'units/s';

    foreach (@$var_strs_ref)
    {
        $var_hash{$_} = $ {$numbers[0]}[$counter];
        $counter++;
    }

    $length = $var_hash{"message length"};
    $time   = $numbers[1];

    return "" if ($time == 0);

    $bandwidth = &$bdw_func_ref ($length, $time, \%var_hash, $param);

    # adjustment of length and bandwidth (<1024 units => units else `Kunits')

    if ($length >= $min_bytes)
    {
        $length     /= 1024;
        $length_str  = '* 1024 units';
    }

    if ($bandwidth >= $min_bytes)
    {
        $bandwidth /= 1024;
        $bdw_str    = '* 1024 units/s';
    }

    return sprintf ("\\item Latency at msg. length %g %s: %g \$\\mu\$s\n".
                    "\\item %s at msg. length %g %s: %g %s",
                    $length, $length_str, $time,
                    ($pattern eq "Collective" ? "Agglomerated bandwidth" :
                     "Bandwidth"), $length, $length_str,
                    $bandwidth, $bdw_str);
}
#### eo comp_bdw ####

####
# comp_bdw_p2p,
# comp_bdw_mw,
# comp_bdw_col_sng, comp_bdw_col_dbl, comp_bdw_col_all
#
# Compute the bandwidth for a measurement in units/s with a certain
# factor determined by the function used.
# The number 1,000,000 is the conversion from units/microsec. to units/sec.
#
# PARAMS:
#   Number    - message length in respective units
#   Number    - time of measurement in microsec.
#   Hash ref. - hash which maps variation params to their respective values
#   Number    - number of nodes/chunks (optional - none with Point-to-Point)
# RETURNS:
#   bandwidth in units/s
####
sub comp_bdw_p2p
{
    my ($length, $time) = @_;

    return (1000000 * 2 * $length / $time);
}

sub comp_bdw_mw
{
    my ($length, $time, $var_hash_ref, $chunks) = @_;

    $chunks = $$var_hash_ref{"chunks"} if (!defined $chunks);

    return (1000000 * $chunks * $length / $time);
}

sub comp_bdw_col_sng
{
    my ($length, $time, $var_hash_ref, $nodes) = @_;

    $nodes = $$var_hash_ref{"number of nodes"} if (!defined $nodes);

    return (1000000 * ($nodes-1) * $length / $time);
}

sub comp_bdw_col_dbl
{
    my ($length, $time, $var_hash_ref, $nodes) = @_;

    $nodes = $$var_hash_ref{"number of nodes"} if (!defined $nodes);

    return (1000000 * 2 * ($nodes-1) * $length / $time);
}

sub comp_bdw_col_all
{
    my ($length, $time, $var_hash_ref, $nodes) = @_;

    $nodes = $$var_hash_ref{"number of nodes"} if (!defined $nodes);

    return (1000000 * $nodes * ($nodes-1) * $length / $time);
}
#### eo comp_bdw_... ####

######## eo bandwidth computation functions ########

####
# add_optional
#
# Creates a comparison section based on a line in the `.dorep' file.
#
# PARAMS:
#   String - title of the section to be created
#   List   - list of filenames of `.ski(c)' files of suites to be compared
# RETURNS:
#   0 (false) if too few suites are comparable or if first suite is not part
#             of main machine
#   1 (true)  if the section was successfully created
####
sub add_optional
{
    my ($title, @file_list) = @_;
    my @headers             = ();  # list of refs to header lists
    my @main_header         = ();  # header list of main machine
    my $ski_file;

    # find common denominator of suites' characteristics
    foreach $ski_file (@file_list)
    {
        if (! -f $ski_file)
        {
            # skip if file hasn't been created
            myprint("Couldn't find `$ski_file' for comparison\n");

            next;
        }

        if (!@main_header)
        {
            if ($ski_file =~ /skic$/)
            {
#                myprint("First suite to be compared with has to be part of ".
#                        "the main machine\n");
#
#                return 0;
## Why not putting an additional machine in front? Maybe someday the whole
## "main machine" idea will be gone anyway...
            }

            @main_header = read_header($ski_file);

            # push main header into header list; this header mustn't be
            # changed afterwards, e. g. its ranges
            my @saved_header = read_header($ski_file);
            push(@headers, \@saved_header);
        }
        else
        {
            my @header = read_header($ski_file);

            ## check if comparable and fit parameters if need be

            # check if variation parameters are compatible
            if (join("_", @{$header[3]}) ne join("_", @{$main_header[3]}))
            {
                myprint("Skipping `$ski_file' because variation parameters ".
                        "aren't compatible\n");

                next;
            }

            # check if scales are compatible
            my $count = 0;
            my $skip  = 0;
            foreach (@{$main_header[6]})
            {
                # only first two params which are relevant for graphics
                last if ($count > 1);

                if (/logarithmic/ && ! $ {$header[6]}[$count]=~/logarithmic/)
                {
                    myprint("Skipping `$ski_file' because scales ".
                            "aren't compatible\n");
                    $skip = 1;

                    last;
                }

                $count++;
            }
            next if ($skip);

            # maybe extend ranges
            $count = 0;
            foreach (@{$main_header[8]})  # range start values
            {
                my $current_value = $ {$header[8]}[$count];

                $_ = $current_value if ($current_value < $_);

                $count++;
            }
            $count = 0;
            foreach (@{$main_header[9]})  # range end values
            {
                my $current_value = $ {$header[9]}[$count];

                $_ = $current_value if ($current_value > $_);

                $count++;
            }

            push(@headers, \@header);

            ##
        } # check parameters
    }  # find common denominator

    # are there any suites left to compare?
    if (@headers < 2)
    {
        myprint("Not enough suites to compare for section `$title'\n");

        return 0;
    }

    my $gpl_file = add_optional_gpl(\@main_header, \@headers);

    local $wide_nt = 0;  # can be set to true in add_optional_node_gpl
    my    $nt_ref  = add_optional_node_gpl(\@main_header, \@headers);

    if (defined($gpl_file))
    {
        system("gnuplot $gpl_file.gpl >$gpl_file.eps")
          && quit("Cannot execute gnuplot with command-file `$gpl_file.gpl'",
                  __LINE__);
        push(@eps_files, "$gpl_file.eps");
    }

    foreach (@$nt_ref)
    {
        system("gnuplot $_.gpl >$_.eps")
          && quit("Cannot execute gnuplot with command-file `$_.gpl'",
                  __LINE__);
        push(@eps_files, "$_.eps");
    }

    add_optional_tex($title, \@main_header, \@headers,
                     $gpl_file, $nt_ref);

    return 1;
}
#### eo add_optional ####

####
# add_optional_tex
#
# Adds a comparison section in the main LaTeX file.
#
# PARAMS:
#   String    - title of the section
#   Reference - ref. to list of information of main suite
#   Reference - ref. to list of list refs of all suites' information
#   String    - filename of gnuplotted file (without extension)
#   Reference - ref. to list of gnuplotted node times files
# RETURNS:
#   1
####
sub add_optional_tex
{
    my ($title,
        $main_header_ref,
        $headers_ref,
        $gpl_file,
        $nt_ref)   = @_;
    my ($ski_file,
        $name,
        $pattern_str,
        $var_strs_ref,
        $unit_strs_ref,
        $format_strs_ref,
        $scale_strs_ref,
        $refine_strs_ref,
        $starts_ref,
        $ends_ref,
        $steps_ref,
        $def_values,
        $standard_error,
        $cut_quantile)   = @$main_header_ref;
    my @headers          = @$headers_ref;
    my $quoted_title     = convert2tex($title);

    if (!$flag++)  # only print at the beginning and only when necessary
    {
        print TEX_FILE <<'EOT';
\section{Comparisons}
  This section includes comparisons of some measurements.
EOT
        if (@comp_info_hash_list)
        {
            $subsection = "subsubsection";
            print TEX_FILE <<'EOT';
  \subsection{Additional machines}
    Here the additional machines are listed with which the measurements of the
    main machine are compared. In section \ref{comp} these machines are
    referred to by numbers in parentheses being added to the suites' names.
    The numbers are the numbers of the machines in this section. When there's
    no number, it's a suite of the main machine.
EOT
            my $count = 0;  # ' (closing quote for my perl mode)
            foreach (@comp_info_hash_list)
            {
                my %hash = %$_;
                my $abs  = "absolute";

                $abs = "not $abs" if ($hash{"ABSOLUTE"} ne "yes");

                foreach (qw(MACHINE INSTITUTION NETWORK NODES NODE DATE))
                {
                    $hash{$_} = '?' unless (exists($hash{$_}));
                }

                $count++;

                print(TEX_FILE <<"EOT");
    \\subsubsection{Machine $count}
    \\label{mach$count}
      \\begin{tabular}{|l|l|}
        \\hline
        Machine & $hash{"MACHINE"} at $hash{"INSTITUTION"}\\\\
        Network & $hash{"NETWORK"} with $hash{"NODES"} nodes\\\\
        Node    & $hash{"NODE"}\\\\
        Results & are $abs\\\\
        Date    & $hash{"DATE"}\\\\
EOT

                foreach (qw(BASETYPE1 BASETYPE2 BASETYPE3 BASETYPE4 BASETYPE5
                            BASETYPE6 BASETYPE7 BASETYPE8 BASETYPE9
                            BASETYPE10))
                {
                    my $keyword = substr($_, 0, 1).lc(substr($_, 1));

                    print(TEX_FILE "        $keyword & $hash{$_}\\\\\n")
                      if (defined($hash{$_}));
                }
                foreach (qw(BLOCKSIZE BLOCKS VECTORSTRIDE))
                {
                    my $keyword = substr($_, 0, 1).lc(substr($_, 1));

                    print(TEX_FILE "        $keyword & $hash{$_}\\\\\n")
                      if (defined($hash{$_}));
                }

                print(TEX_FILE <<"EOT");
        \\hline
      \\end{tabular}
EOT
            }

            print TEX_FILE <<'EOT';
  \subsection{Comparisons}
  \label{comp}
EOT
        }  # other machines exist
    }

    print(TEX_FILE <<"EOT");
\\${subsection}{$quoted_title}
EOT

    if ($pattern_str ne "Simple")
    {
        if (defined($gpl_file))
        {
            print(TEX_FILE <<"EOT");
%\\begin{figure}[Htb]
  \\begin{center}
     \\epsfig{file=$gpl_file.eps}
  \\end{center}
  %\\caption{$quoted_title.}
%\\end{figure}
EOT
        }
        else
        {
            # create table
            my $create_output = 0;

            foreach (@headers)
            {
                my $ski_file = $$_[0];
                my $suite    = "$$_[2] $$_[1]";
                my @data     = read_data($ski_file);
                my $num_col  = 3 + @$var_strs_ref;

                next if (@data >= $minimum_data);

                if (!$create_output)
                {
                    my $arguments = join('} & \\textbf{', @$var_strs_ref);

                    print(TEX_FILE "\\begin{center}\n  \\begin{tabular}{|".
                          ('r|' x $num_col)."}\n    \\hline\n");
                    print(TEX_FILE "    \\textbf{$arguments} & ".
                          "\\textbf{Result} & \\textbf{Std. Error} & ".
                          "\\textbf{Count}\\\\\n    \\hline\n");

                    $create_output = 1;
                }

                $suite .= " ($1)" if ($ski_file =~ /(\d+)\.skic$/);
                $suite  = convert2tex($suite);

                print(TEX_FILE "    \\hline\n    ".
                      "\\multicolumn{$num_col}{|c|}{\\textbf{$suite}}\\\\\n".
                      "    \\hline\n");

                foreach (@data)
                {
                    print(TEX_FILE '    '.join(' & ', @{$$_[0]}, @$_[1..3]).
                          "\\\\\n");
                }

                print(TEX_FILE "    \\hline\n");
            }

            print(TEX_FILE "  \\end{tabular}\n\\end{center}\n")
              if ($create_output);
        }

        if (@$nt_ref)
        {
            my $eps_count = 0;

            # allow page break between wide eps files (s. header2tex_module)
            print(TEX_FILE "  \\begin{center}\n") unless ($wide_nt);

            foreach (@$nt_ref)
            {
                # end centering after every second small image
                # to allow page break there
                if (!$wide_nt && $eps_count && !($eps_count % 2))
                {
                    print(TEX_FILE "  \\end{center}\n");
                    print(TEX_FILE "  \\begin{center}\n");
                }

                print(TEX_FILE "  \\begin{center}\n") if ($wide_nt);

                print TEX_FILE <<"EOT";
    \\epsfig{file=$_.eps}
EOT

                print(TEX_FILE "  \\end{center}\n") if ($wide_nt);

                $eps_count++;
            }

            print(TEX_FILE "  \\end{center}\n") unless ($wide_nt);
        }  # eps files for node times

        print(TEX_FILE create_comp_table($headers_ref));

        # print suites' pieces of information
        print(TEX_FILE <<"EOT");
  \\begin{itemize}
EOT
        foreach (@headers)
        {
            my $suite = "$$_[2] $$_[1]";

            $suite .= " ($1)" if ($$_[0] =~ /(\d+)\.skic$/);
            $suite  = convert2tex($suite);

            print(TEX_FILE <<"EOT");
    \\item Information for suite `$suite'
EOT
            print_suite_info(TEX_FILE, @$_);  # ` (for my perl mode)
        }
        print(TEX_FILE <<"EOT");
  \\end{itemize}
\\newpage
EOT
    }  # not simple
    else
    {
        print(TEX_FILE <<EOT);
\\begin{center}
  \\begin{tabular}{|r|r|r|}
    \\hline
    Result [\$\\mu\$s] & Standard error [\$\\mu\$s] & Count\\\\
    \\hline
EOT
        foreach (@headers)
        {
            my $ski_file  = $$_[0];
            my $suite     = "$$_[2] $$_[1]";
            my @data_list = read_data($ski_file);

            $suite .= " ($1)" if ($ski_file =~ /(\d+)\.skic$/);
            $suite  = convert2tex($suite);

            print(TEX_FILE "    \\hline\n    ".
                  "\\multicolumn{3}{|c|}{\\textbf{$suite}}\\\\\n".
                  "    \\hline\n");

            foreach (@data_list)
            {
                my ($result, $std_error, $count) = @$_[1..3];

                print(TEX_FILE "$result & $std_error & $count\\\\\n");
            }

            print(TEX_FILE "    \\hline\n");
        }

        print(TEX_FILE <<EOT);
  \\end{tabular}
\\end{center}
EOT

        print(TEX_FILE <<EOT);
\\begin{itemize}
EOT

        foreach (@headers)
        {
            my $ski_file  = $$_[0];
            my $suite     = "$$_[2] $$_[1]";
            my $pattern   = $$_[2];
            my $std_err   = $$_[12];

            $suite .= " ($1)" if ($ski_file =~ /(\d+)\.skic$/);
            $suite  = convert2tex($suite);

            print(TEX_FILE <<EOT);
  \\item Information for suite `$suite'
  \\begin{itemize}
EOT
            print(TEX_FILE <<EOT);  # ` (for my perl mode)
    \\item Pattern: $pattern
    \\item Max. allowed standard error is $std_err \\%
  \\end{itemize}
EOT
        }

        print(TEX_FILE <<EOT);
\\end{itemize}
EOT
    }

    return 1;
}
#### eo add_optional_tex ####

####
# add_optional_gpl
#
# Produces a gnuplot file which gnuplot uses to produce an `.eps' file
# which shows the suites' data.
#
# PARAMS:
#   Reference - ref. to list of information of main suite (s. read_header)
#   Reference - ref. to list of list refs of all suites' information
# RETURNS:
#   String - filename (without extension) to be gnuplotted (or undef)
####
sub add_optional_gpl
{
    my ($main_header_ref,
        $headers_ref)     = @_;
    my @main_header       = @$main_header_ref;
    my @var_strs          = @{$main_header[3]};
    my @scales            = @{$main_header[6]};
    my @starts            = @{$main_header[8]};
    my @ends              = @{$main_header[9]};
    my @headers           = @$headers_ref;
    my $file_name         = sprintf("$pattern_str_hash{$main_header[2]}_".
                                    "$main_header[1].comp%.3d", $file_count);
    my $enough_data       = 0;
    my $header_ref;

    # check if maximum number of data lines is enough for eps creation
    foreach $header_ref (@headers)
    {
        my $ski = $$header_ref[0];
        my @data_list = read_data($ski);

        if (@data_list >= $minimum_data)
        {
            $enough_data = 1;

            last;
        }
    }
    # not enough data lines for `.eps' file? (Simple pattern has just 1 data
    # line, so no image is produced (s. also -m option))
    return undef unless ($enough_data);

    if (@var_strs > 2)
    {
        myprint("NOTE: no graphical output supported for ".
                scalar(@var_strs)."-dimensional data\n");

        return undef;  # can't create an eps file
    }

    $file_count++;

    if (-f "$file_name.gpl" && !$overwrite)
    {
        myprint("Using existing gnuplot file `$file_name.gpl'\n");

        return $file_name;
    }

    open(GPL_FILE,">$file_name.gpl")
      || quit("Cannot open `$file_name.gpl': $!", __LINE__);
    push(@files_to_delete, "$file_name.gpl");

    my $output3d = "";
    $output3d = "set parametric\nset dgrid3d" if (@var_strs == 2);

    my $label_str = "set xlabel \"$var_strs[0]\"\n".
      "set ylabel ";
    if (@var_strs == 2)
    {
        $label_str .= "\"$var_strs[1]\"\n".
          "set zlabel ";
    }
    $label_str .= '"time (microsec.)"';

    my $x_log_scale = "";
    my $y_log_scale = "";
    my $z_log_scale = "";
    if ($scales[0] =~ /logarithmic/)
    {
        $x_log_scale = "set logscale x 10";
        $y_log_scale = "set logscale y 2";
        $z_log_scale = "set logscale z 2";
    }
    if (@var_strs == 2)
    {
        if ($scales[1] =~ /logarithmic/)
        {
            $y_log_scale = "set logscale y 10";
            $z_log_scale = "set logscale z 2";
        }
        else
        {
            $y_log_scale = "";
        }
    }

    print(GPL_FILE <<EOT);
#gnuplot command-file for the measurement $main_header[1]
set nologscale
$output3d
set xtics
set ytics
set ztics
set term postscript eps 18
$label_str
$x_log_scale
$y_log_scale
$z_log_scale
set output  # to be redirected to $file_name.eps
EOT
    if (@var_strs == 1)
    {
        print(GPL_FILE "plot [$starts[0]:$ends[0]]");
    }
    else
    {
        print(GPL_FILE "splot [$starts[0]:$ends[0]] [$starts[1]:$ends[1]]");
    }

    my $first = 1;
    foreach (@headers)
    {
        my $ski_file = $$_[0];
        my $name     = $$_[2]." ".$$_[1];

        $name .= " ($1)" if ($ski_file =~ /(\d+)\.skic$/);

        print(GPL_FILE ",") if (!$first);
        $first = 0;

        print(GPL_FILE " \"$ski_file.cut\" title \"$name\" with ");
        if (@var_strs == 1)
        {
            print(GPL_FILE "linesp");
        }
        else
        {
            print(GPL_FILE "impulses");
        }
    }

    close(GPL_FILE);

    return $file_name;
}
#### eo add_optional_gpl ####

####
# add_optional_node_gpl
#
# Produces gnuplot files which gnuplot uses to produce `.eps' files which
# show a node times comparison of the suites.
#
# PARAMS:
#   Reference - ref. to list of information of main suite (s. read_header)
#   Reference - ref. to list of list refs of all suites' information
# RETURNS:
#   Reference - ref. to list of filenames (without extension) to be gnuplotted
####
sub add_optional_node_gpl
{
    my ($main_header_ref,
        $headers_ref)     = @_;
    my ($ski_file,
        $name,
        $pattern_str,
        $var_strs_ref,
        $unit_strs_ref,
        $format_strs_ref,
        $scale_strs_ref,
        $refine_strs_ref,
        $starts_ref,
        $ends_ref,
        $steps_ref,
        $def_values,
        $standard_error,
        $cut_quantile)   = @$main_header_ref;
    my @headers          = @$headers_ref;
    my $node_file        = "";  # filename of gnuplot file for node times
    my $data_count       = -1;  # counter for data lines
    my @to_be_created    = ();  # files to be created (returned)

    my @data_list = read_data($ski_file);

    # we don't need the main suite here
    shift(@headers);

    # examine if node times should be shown in an `.eps' file
    foreach (@data_list)
    {
        my @args            = @{$$_[0]};
        my @node_times      = @{$$_[7]};
        my $title_var_str   = "";
        my @legends         = ();  # legends for the suites
        my @node_times_list = ();  # node times for the suites (list of lists)
        my $legend          = "";  # single legend

        $data_count++;

        # do node times actually exist?
        next unless (@node_times);

        my $arg_count = 0;  # number of arguments which fit for barchart
        foreach (@$var_strs_ref)
        {
            my $var_str = $_;
            (my $var_str_ = $_) =~ s/ /_/g;

            if (!grep("none" eq $_, @{$barchart_values{$var_str_}}))
            {
                foreach (@{$barchart_values{$var_str_}})
                {
                    if (($_ eq "min" && $data_count == 0) ||
                        ($_ eq "max" && $data_count == $#data_list) ||
                        ($_ ne "min" && $_ ne "max" &&
                         $_ == $args[$arg_count]))
                    {
                        $title_var_str .= ", " if ($title_var_str ne "");
                        $title_var_str .= "$var_str $args[$arg_count]";
                        $arg_count++;
                        last;
                    }
                }
            }
        }

        my $max_nodes = @node_times;

        # test (sorted) node times for all of them being 0
        next unless ($node_times[0] > 0);

        $legend  = "$pattern_str $name";
        $legend .= " ($1)" if ($ski_file =~ /(\d+)\.skic$/);

        push(@legends,         $legend);
        push(@node_times_list, \@node_times);

        # all arguments designated for barchart output?
        if ($arg_count == @$var_strs_ref)
        {
            my $header_ref;
            my $image_size     = 'set size     .5, .5';
            my $num_node_times = 0;

            # determine the other suites where the same arguments appear
            foreach $header_ref (@headers)
            {
                my @add_data_list = read_data($$header_ref[0]);

                foreach (@add_data_list)
                {
                    my @comp_args = @{$$_[0]};
                    my @comp_nt   = @{$$_[7]};

                    # do node times exist?
                    next unless (@comp_nt);

                    # check if arguments are the same
                    next if (join("_", @comp_args) ne join("_", @args));

                    # test (sorted) node times for all of them being 0
                    next unless ($comp_nt[0] > 0);

                    # adjust number of node times
                    $max_nodes = @comp_nt if (@comp_nt > $max_nodes);

                    # create legend to this suite
                    my $legend = "$$header_ref[2] $$header_ref[1]";
                    $legend .= " ($1)" if ($$header_ref[0] =~ /(\d+)\.skic$/);

                    push(@legends,         $legend);
                    push(@node_times_list, \@comp_nt);

                    # next suite's header
                    last;
                }
            }

            my $number_of_suites = @legends;

            # next data line if nothing to compare
            next if ($number_of_suites < 2);

            $node_file = sprintf("$pattern_str_hash{$pattern_str}_$name".
                                 ".comp.nt%.3d", $file_count);

            # calculate values for positioning in eps file
            my $boxwidth  = 1 / ($number_of_suites + 1 +
                                 ($number_of_suites-1) / 2);
            my $stepwidth = 1.5 * $boxwidth;
            my $pos       = $boxwidth;

            foreach (0..$#legends)
            {
                my $suite_file = sprintf("$node_file.%.2d", $_);

                open(SUITE_FILE, ">$suite_file")
                  || quit("Cannot open `$suite_file': $!", __LINE__);
                push(@files_to_delete, $suite_file);

                # write sorted node times to file
                my $node_count = .5 + $pos;
                foreach (@{$node_times_list[$_]})
                {
                    print(SUITE_FILE "$node_count $_\n");
                    $node_count++;
                }

                close(SUITE_FILE);

                $pos += $stepwidth;
            }

            push(@to_be_created, $node_file);

            $file_count++;

            if (-f "$node_file.gpl" && !$overwrite)
            {
                myprint("Using existing gnuplot file `$node_file.gpl'\n");

                next;
            }

            open(GPL_FILE, ">$node_file.gpl")
              || quit("Cannot open `$node_file.gpl': $!", __LINE__);
            push(@files_to_delete, "$node_file.gpl");

            # produce wider image if lots of node times will be plotted
            foreach (@node_times_list)
            {
                $num_node_times += @$_;
            }

            if ($num_node_times >= $min_node_times)
            {
                $image_size = 'set size     1, .5';

                $wide_nt = 1;  # for add_optional_tex to allow page break
            }

            $max_nodes += .5;  # range end for gpl file

            print(GPL_FILE <<"EOT");
# gnuplot file for node times of measurement $name
set term     postscript eps 10
$image_size
set nogrid  #   ytics
set xtics    1, 1
set boxwidth $boxwidth
set xlabel   "node number (sorted)"
set ylabel   "time (microsec.)"
set title    "node times for $title_var_str"
set output  # to be redirected to $node_file.eps
EOT
            print(GPL_FILE "plot [.5:$max_nodes]");

            my $first = 1;

            foreach (0..$#legends)
            {
                my $suite_file = sprintf("$node_file.%.2d", $_);

                print(GPL_FILE ",") if (!$first);
                $first = 0;

                print(GPL_FILE " \"$suite_file\" title \"$legends[$_]\" ".
                      "with boxes");
            }

            print(GPL_FILE "\n");

            close(GPL_FILE);
        }  # create barchart
    }  # loop over data lines

    return \@to_be_created;
}
#### eo add_optional_node_gpl ####

####
# create_comp_table
#
# Creates a string with LaTeX output of a table comparing the bandwidths of
# the compared suites.
#
# PARAMS:
#   Reference - ref. to list of suites' headers
# RETURNS:
#   String - comparison LaTeX table or empty if no message lengths available
####
sub create_comp_table
{
    my $headers_ref  = shift @_;
    my @headers      = @$headers_ref;
    my @var_strs     = @{$ {$headers[0]}[3]};
    my $param_str    = "\\multicolumn{3}{|c||}{(other params also min.)} & ".
                       "\\multicolumn{3}{c|}{(other params also max.)}\\\\";
    my $other_params = "";
    my $header_str;
    my $middle_str   = "";
    my $trailer_str;
    my $min_msg_len_min  = undef;
    my $min_msg_len_unit = 'units';
    my $min_bdw_min      = undef;
    my $min_bdw_unit     = 'units/s';
    my $max_msg_len_min  = undef;
    my $max_msg_len_unit = 'units';
    my $max_bdw_min      = undef;
    my $max_bdw_unit     = 'units/s';
    my @table_values = ();  # matrix of values per suite

    $other_params = $param_str if (@var_strs > 1);

    foreach (@headers)
    {
        my $ski_file     = $$_[0];
        my $name         = $$_[1];
        my $pattern_str  = $$_[2];
        my $var_strs_ref = $$_[3];
        my $def_values   = $$_[11];
        my $suite_texed  = "$pattern_str $name";

        $suite_texed .= " ($1)" if ($ski_file =~ /(\d+)\.skic$/);
        $suite_texed  = convert2tex($suite_texed);

        if (grep("message length" eq $_, @$var_strs_ref)
            || $def_values =~ /message length (\d+)/)
        {
            my $default_length = $1;
            my @data_list      = read_data($ski_file);
            my @min_max_lines  = ();
            my @suite_values   = ($suite_texed);  # + values for table row
            my $max_part       = 0;  # min or max part in foreach loop below?

            push (@min_max_lines, $data_list[0]);
            push (@min_max_lines, $data_list[-1]);

            foreach (@min_max_lines)
            {
                my @args        = @{$$_[0]};
                my $length      = 0;
                my $time        = $$_[1];
                my %var_hash    = ();
                my $counter     = 0;
                my $len_min_ref = ($max_part) ? \$max_msg_len_min
                                              : \$min_msg_len_min;
                my $bdw_min_ref = ($max_part) ? \$max_bdw_min
                                              : \$min_bdw_min;

                foreach (@$var_strs_ref)
                {
                    $var_hash{$_} = $args[$counter];
                    $counter++;
                }

                if (defined $var_hash{"message length"})
                {
                    $length = $var_hash{"message length"};
                }
                else
                {
                    $length       = $default_length;
                    $other_params = $param_str if ($other_params eq "");
                }

                push (@suite_values, $length, $time);

                $$len_min_ref = $length if (!defined ($$len_min_ref) ||
                                            $length < $$len_min_ref);

                if ($time == 0)
                {
                    push (@suite_values, "---");
                }
                else
                {
                    my ($bdw_func_ref,
                        $param)        = determine_bdw_func ($ski_file,
                                                             $name,
                                                             $pattern_str,
                                                             $var_strs_ref,
                                                             $def_values);

                    if (defined $bdw_func_ref)
                    {
                        my $bandwidth = &$bdw_func_ref ($length,
                                                        $time,
                                                        \%var_hash,
                                                        $param);
                        push (@suite_values, $bandwidth);

                        $$bdw_min_ref = $bandwidth
                          if (!defined ($$bdw_min_ref) ||
                              $bandwidth < $$bdw_min_ref);
                    }
                    else
                    {
                        push (@suite_values, "?");
                    }
                }

                $max_part = 1;
            }  # min. / max. data

            push (@table_values, \@suite_values);
        }
    }  # headers

    # create middle_str with adjusted values
    foreach (@table_values)
    {
        my @suite_values = @$_;
        my $suite_texed  = shift(@suite_values);

        # convert message lengths and bandwidths into multiples of 1024
        # units(/s) if appropriate
        if ($min_msg_len_min >= $min_bytes)
        {
            $suite_values[0] /= 1024;
            $min_msg_len_unit  = '1024 units';
        }
        if ($max_msg_len_min >= $min_bytes)
        {
            $suite_values[3] /= 1024;
            $max_msg_len_unit  = '1024 units';
        }
        if ($min_bdw_min >= $min_bytes)
        {
            $suite_values[2] /= 1024;
            $min_bdw_unit  = '1024 units/s';
        }
        if ($max_bdw_min >= $min_bytes)
        {
            $suite_values[5] /= 1024;
            $max_bdw_unit  = '1024 units/s';
        }

        $middle_str .= "\\multicolumn{6}{|c|}{$suite_texed}\\\\ \\hline";
        $middle_str .= sprintf(" %d & %g & %g & %g & %g & %g\\\\".
                               "\\hline\n", @suite_values);
    }

    return "" if ($middle_str eq "");

    $header_str = <<"EOT";
\\begin{center}
\\begin{tabular}{|r|r|r||r|r|r|}
\\hline
msg. len. & latency  & bandwidth & msg. len & latency  & bandwidth\\\\
{[$min_msg_len_unit]}  & [\$\\mu\$s] & [$min_bdw_unit] &
[$max_msg_len_unit]    & [\$\\mu\$s] & [$max_bdw_unit]\\\\
\\hline
\\multicolumn{3}{|c||}{Min. msg. length} & \\multicolumn{3}{c|}{Max. msg. length}\\\\
$other_params
\\hline
EOT
    $trailer_str = <<'EOT';
\end{tabular}
\end{center}
EOT

    return ($header_str.$middle_str.$trailer_str);
}
#### eo create_comp_table ####

####
# check_environment
#
# Checks the environment variables for options for this script.
# At the moment there's just the possibility to adjust the name of the
# `.dorep' file with the variable DOREP. This has to be done this way
# because the options in the `.dorep' file should be extracted before
# the command line options, so the user wouldn't be able to specify a
# name for the file other than `.dorep'.
#
# PARAMS:
#   none
# RETURNS:
#   nothing
####
sub check_environment
{
    if (defined $ENV{"DOREP"})
    {
        $dorep_opt_file = $ENV{"DOREP"};
    }
}
#### eo check_environment ####

####
# parse_parameters
#
# Reads the command line options and sets the respective global variables.
#
# PARAMS:
#   none
# RETURNS:
#   nothing
####
sub parse_parameters
{
    my $iteration = 0;  # position in @ARGV of next option
    my $skip      = 0;  # number of options to skip if already parsed

    foreach (@ARGV)
    {
        $iteration++;

        if ($skip)
        {
            $skip--;
            next;
        }

        # sort of switch
        (/^-b$/ ||
         /^--barchart$/) &&
           do
           {
               if ($iteration <= $#ARGV)
               {
                   my @entries = split (/\s*;\s*/, $ARGV[$iteration]);
                   foreach (@entries)
                   {
                       if (/(.+?)\s+(.+)/)
                       {
                           my $hash_key = $1;
                           if (!grep($hash_key eq $_, @known_parameters))
                           {
                               mywarn("Variation parameter `$hash_key' ".
                                      "in barchart entry\n\t`$_'\n".
                                      "at command line ".
                                      "unknown by the time this ".
                                      "script was written.\n");
                           }
                           my @hash_values = split /\s*,\s*/, $2;
                           $barchart_values{$hash_key} = \@hash_values;
                       }
                       else
                       {
                           mywarn("Wrong barchart entry `$_' on ".
                                  "command line.\n");
                       }
                   }
               }
               else
               {
                   print_usage_quit("Missing barchart info with option $_");
               }
               $skip = 1;
               next;
           };
        (/^-d$/ ||
         /^--dri_directory$/) &&
           do
           {
               if ($iteration <= $#ARGV)
               {
                   unless (-d $ARGV[$iteration])
                   {
                       mywarn("Directory `$ARGV[$iteration]' (dri-files) ".
                              "not found. Trying default directory.\n");
                       $skip = 1;
                       next;
                   }
                   $dri_directory = $ARGV[$iteration];
                   # add '/' if need be
                   if ($dri_directory !~ /\/$/)
                   {
                       $dri_directory .= "/";
                   }
               }
               else
               {
                   print_usage_quit("Missing pathname with option $_");
               }
               $skip = 1;
               next;
           };
        (/^-dvi$/ ||
         /^--output_dvi$/) &&
           ($to_dvi = 1, next);  # undocumented option
        (/^-h$/ ||
         /^--help$/) &&
           print_help_quit();
        (/^-m$/ ||
         /^--minimum_data$/) &&
           do
           {
               if ($iteration <= $#ARGV)
               {
                   $minimum_data = $ARGV[$iteration];
               }
               else
               {
                   print_usage_quit("Missing number with option $_");
               }
               if ($minimum_data < 2)
               {
                   mywarn("Invalid value for option $_:\nAt least 2 data ".
                          "points necessary for graphical output. Setting ".
                          "value to 2.\n");

                   $minimum_data = 2;
               }
               $skip = 1;
               next;
           };
        (/^-p$/ ||
         /^--output$/) &&
           do
           {
               if ($iteration <= $#ARGV)
               {
                   push(@redirect, $ARGV[$iteration]);
               }
               else
               {
                   print_usage_quit("Missing filename with option $_");
               }
               $skip = 1;
               next;
           };
        (/^-q$/ ||
         /^--quiet$/) &&
           ($quietness = $quiet, next);
        (/^-Q$/ ||
         /^--very_quiet$/) &&
           ($quietness = $very_quiet, next);
        (/^-t$/ ||
         /^--keep_temp$/) &&
           ($wastefulness = $delete_nothing, next);
        (/^-T$/ ||
         /^--delete_temp_on_error$/) &&
           ($wastefulness = $delete_temp_on_error, next);
        (/^-tex$/ ||
         /^--output_tex$/) &&
           ($to_tex = 1, next);  # undocumented option
        (/^-u$/ ||
         /^--use_temp$/) &&
           ($overwrite = 0, next);
        /^-/ &&
          print_usage_quit("Unknown option: $_");
        /(.*)/ &&
          ($skampi_data_file = $1, next);
    }
}
#### eo parse_parameters ####

####
# read_options
#
# Reads the option part of the option file (`.dorep' by default) and
# sets the respective global variables.
#
# PARAMS:
#   none
# RETURNS:
#   nothing
####
sub read_options
{
    my $in_options_part     = 0;
    my $in_suites_part      = 0;
    my $in_exclude_part     = 0;
    my $in_comparisons_part = 0;

    if (open (OPT_FILE, "<$dorep_opt_file"))
    {
        my $saved_arg = $_;  # $_ isn't recovered after while (<...>)
        while (<OPT_FILE>)
        {
            chomp();

            next if (/^\#/ || /^\s*$/);

            if (/^\[(.+?)\]/)
            {
                $in_options_part     = 0;
                $in_suites_part      = 0;
                $in_exclude_part     = 0;
                $in_comparisons_part = 0;

                if (lc($1) eq 'options')
                {
                    $in_options_part = 1;
                }
                elsif (lc($1) eq 'suites')
                {
                    $in_suites_part    = 1;
                    $report_all_suites = 0;   # even if this section is empty
                    @excluded_suites   = ();  # sections are exclusive
                }
                elsif (lc($1) eq 'exclude')
                {
                    $in_exclude_part   = 1;
                    $report_all_suites = 1;
                    @reported_suites   = ();  # sections are exclusive
                }
                elsif (lc($1) eq 'comparisons')
                {
                    $in_comparisons_part = 1;
                }
                else
                {
                    quit("Unknown section in line $. of parameter file ".
                         "`$dorep_opt_file'");
                }
            }
            elsif ($in_options_part ||
                   (!$in_suites_part      && !$in_exclude_part &&
                    !$in_comparisons_part && /^-/))                # old format
            {
                # sort of switch
                (/^-b\s+(.+?)\s*$/ ||
                 /^--barchart\s+(.+?)\s*$/) &&
                   do
                   {
                       my @entries = split /\s*;\s*/, $1;
                       foreach (@entries)
                       {
                           if (/(.+?)\s+(.+)/)
                           {
                               my $hash_key = $1;
                               if (!grep($hash_key eq $_, @known_parameters))
                               {
                                   mywarn ("Variation parameter `$hash_key' ".
                                           "in barchart entry\n\t`$_'\n".
                                           "in parameter file ".
                                           "`$dorep_opt_file' ".
                                           "unknown by the time this ".
                                           "script was written.\n");
                               }
                               my @hash_values = split /\s*,\s*/, $2;
                               $barchart_values{$hash_key} = \@hash_values;
                           }
                           else
                           {
                               mywarn ("Wrong barchart entry `$_' in ".
                                       "parameter file ".
                                       "`$dorep_opt_file'.\n");
                           }
                       }
                       next;
                   };
                (/^-d\s+(.*)/ ||
                 /^--dri_directory\s+(.*)/) &&
                   do
                   {
                       unless (-d "$1")
                       {
                           mywarn ("Directory `$1' (dri-files) not found. ".
                                   "Trying default directory.\n");

                           next;
                       }
                       $dri_directory = $1;
                       # add '/' if need be
                       if ($dri_directory !~ /\/$/)
                       {
                           $dri_directory .= "/";
                       }
                       next;
                   };
                (/^-dvi$/ ||
                 /^--output_dvi$/) &&
                   ($to_dvi = 1, next);  # undocumented option
                (/^-f\s+(.*)/ ||
                 /^--skampi_file\s+(.*)/) &&
                   ($skampi_data_file = $1, next);
                (/^-m\s+(\d+?)\s*$/ ||
                 /^--minimum_data\s+(\d+?)\s*$/) &&
                   do
                   {
                       $minimum_data = $1;
                       if ($minimum_data < 2)
                       {
                           mywarn("Invalid value in line $. of parameter ".
                                  "file `$dorep_opt_file':\nAt least 2 data ".
                                  "points necessary for graphical output. ".
                                  "Setting value to 2.\n");

                           $minimum_data = 2;
                       }
                       next;
                   };
                (/^-p\s+(.*)/ ||
                 /^--output\s+(.*)/) &&
                   (push (@redirect, $1), next);
                (/^-q/ ||
                 /^--quiet/) &&
                   ($quietness = $quiet, next);
                (/^-Q/ ||
                 /^--very_quiet/) &&
                   ($quietness = $very_quiet, next);
                (/^-t/ ||
                 /^--keep_temp/) &&
                   ($wastefulness = $delete_nothing, next);
                (/^-T/ ||
                 /^--delete_temp_on_error/) &&
                   ($wastefulness = $delete_temp_on_error, next);
                (/^-tex$/ ||
                 /^--output_tex$/) &&
                   ($to_tex = 1, next);  # undocumented option
                (/^-u/ ||
                 /^--use_temp/) &&
                   ($overwrite = 0, next);
                mywarn ("Unknown option `$_' in parameter file ".
                        "`$dorep_opt_file'.\n");
            }
            elsif ($in_suites_part)
            {
                /^\s*(.+?)\s*$/;
                push(@reported_suites, "$1.ski");
            }
            elsif ($in_exclude_part)
            {
                /^\s*(.+?)\s*$/;
                push(@excluded_suites, "$1.ski");
            }
            else  # don't test for being inside comparison section (old format)
            {
                $in_comparisons_part = 1;  # don't parse for options any longer

                if (/^(.+?)\s*:\s*(.+?)\s*$/)
                {
                    my $title     = $1;
                    my @file_list = split(/\s*,\s*/, $2);

                    foreach (@file_list)
                    {
                        # adjust to filename
                        if (/\#\d+$/)
                        {
                            $_ .= '.skic';
                        }
                        else
                        {
                            $_ .= '.ski';
                        }
                    }

                    push(@comparisons,     [$title, \@file_list]);
                    push(@compared_suites, @file_list);
                }
                else
                {
                    mywarn("Error in comparison entry in line $. of ".
                           "parameter file `$dorep_opt_file'\n");
                }
            }
        }
        $_ = $saved_arg;
    }
    else
    {
        mywarn ("Cannot open optional parameter file `$dorep_opt_file' for ".
                "$Script_Name options.\n");
        return;
    }

    close (OPT_FILE);
}
#### eo read_options ####

####
# myprint
#
# Prints a message depending on the selected `quietness'.
#
# PARAMS:
#   List of Strings - Strings to be printed
# RETURNS:
#   return value of print or 0 if quiet
####
sub myprint
{
    if ($quietness == $verbose)
    {
        return print @_;
    }

    return 0;
}
#### eo myprint ####

####
# mywarn
#
# Prints a warning depending on the selected `quietness'.
#
# PARAMS:
#   List of Strings - Strings to be printed
# RETURNS:
#   return value of warn or 0 if very quiet
####
sub mywarn
{
    # WARNING (when changing): comparing with `<'
    if ($quietness < $very_quiet)
    {
        return warn @_;
    }

    return 0;
}
#### eo mywarn ####

####
# print_help_quit
#
# Prints a help message and quits.
#
# PARAMS:
#   none
# RETURNS:
#   nothing (calls `exit' with value 0 (no error))
####
sub print_help_quit
{
    # ignore possibly set output file
    select STDOUT;

    print $usage_header;
    print $usage;
    print $options;

    exit(0);
}
#### eo print_help_quit ####

####
# print_usage_quit
#
# Prints a usage message and quits.
#
# PARAMS:
#   String - additional error message (optional)
# RETURNS:
#   nothing (calls `exit' with value 1)
####
sub print_usage_quit
{
    # ignore possibly set output file
    select STDOUT;

    print $usage_header;
    print "".(shift @_)."\n" if @_;
    print $usage;
    print "Type $Script_Name.pl -h to get further information\n";

    exit(1);
}
#### eo print_usage_quit ####

####
# quit
#
# Deletes the temporary files if the respective variable is set appropriately
# and quits the script with an error message.
# This function should be called with a message string and the line
# number where this function is called (__LINE__). If the string is missing
# a default string is printed. If the line number is missing it defaults
# to the line number where `die' is called in this function.
#
# PARAMS:
#   String - error message (optional)
#   Number - line number (optional)
# RETURNS:
#   nothing (lets script exit with value 1)
####
sub quit
{
    my ($message,
        $line)    = @_;
    my $temp_info = '';

    # delete temporary files if need be
    $temp_info = kill_temp($wastefulness == $delete_temp_on_error);

    close(REDIRECT);

    if (defined($message))
    {
        $message = $message.'. ';
    }
    else
    {
        $message = '';
    }
    if (defined($line))
    {
        warn("${message}`", __FILE__, "' quitted at line $line.\n$temp_info");
    }
    else
    {
        warn("${message}`", __FILE__, "' died.\n$temp_info");
    }

    exit(1);
}
#### eo quit ####

####
# kill_temp
#
# Deletes temporary files or creates scripts with which the user can do this.
# Returns a string which tells the user about these scripts if appropriate.
#
# PARAMS:
#   $ - 0, if temporary files are not to be deleted
#       1, if temporary files are to be deleted
# RETURNS:
#   $ - information string (empty if no scripts were created)
####
sub kill_temp
{
    my $delete    = shift;
    my $temp_info = '';

    push(@files_to_delete, splice(@eps_files, 0)) unless ($to_tex || $to_dvi);

    if (@files_to_delete)
    {
        if ($delete)
        {
            myprint("Deleting temporary files...\n");

            unlink(@files_to_delete);
        }
        else
        {
            if (open(TEMP, '>kill_temp'))
            {
                foreach (@files_to_delete)
                {
                    print(TEMP "rm -f $_\n");
                }
                print(TEMP "rm -f kill_temp\n");

                close(TEMP);
                chmod(0755, 'kill_temp');

                $temp_info = "Temporary files can be deleted by executing ".
                  "the script `kill_temp'.\n";
            }
        }
    }

    if (@eps_files)
    {
        if (open(EPS, '>kill_eps'))
        {
            foreach (@eps_files)
            {
                print(EPS "rm -f $_\n");
            }
            print(EPS "rm -f kill_eps\n");

            close(EPS);
            chmod(0755, 'kill_eps');

            $temp_info .= "EPS files can be deleted by executing ".
              "the script `kill_eps'.\n";
        }
    }

    return $temp_info;
}

####
# convert2tex
#
# Converts a string so that LaTeX is able to handle it. For that purpose it
# just converts special characters.
#
# PARAMS:
#   $ - string
# RETURNS:
#   $ - converted string
####
sub convert2tex
{
    my $string    = shift;
    my $converted = $string;
    my %conv_hash =
        (
         '#'    => '\\#',
         '$'    => '\\$',
         '%'    => '\\%',
         '&'    => '\\&',
         '~'    => '\\symbol{126}',
         '_'    => '\\_',
         '^'    => '\\symbol{94}',
         '\\'   => '$\\backslash$',
         '{'    => '\\{',
         '}'    => '\\}',
        );
    my $conv_chars = join ('', keys (%conv_hash));

    ## Explanation:
    ## In the following substitution the regular expression (without the
    ## explicit backslash) would find a backslash only if a `\\\\' appeared in
    ## the hash table above. That's why the backslash is explicitly specified
    ## in the next line.
    ## The substitution, however, only takes place if `\\' appears in the
    ## hash table.

    $converted =~ s/([\\$conv_chars])/$conv_hash{$1}/g;

    return $converted;
}
#### eo convert2tex ####

####
# parse_run_info
#
# Parses the part of a SKaMPI output file where the information on the SKaMPI
# run is located.
# If the parameter to this function is a file handle, the position within it
# after the function has ended is just before the suites part.
#
# PARAMS:
#   @ - list of data lines
#       (if the first parameter is a file handle (or reference to it), read
#        from that one instead)
# RETURNS:
#   % - hash of run information
#       text after `#@' => text after that text
#                          (multiple lines joined with ` ')
####
sub parse_run_info
{
    my @lines        = @_;
    my $file_hdl     = $lines[0];
    my %run_hash     = ();
    my $key          = undef;
    my $reading_file = 0;      # parameters are data lines (no file handle)
    my $line_nr      = 0;      # number of parsed line (1-based)
    my $line;

    if (ref($lines[0]) eq 'GLOB' || ref(\$lines[0]) eq 'GLOB')
    {
        $reading_file = 1;

        $line = <$file_hdl>;
    }
    else
    {
        $line = shift(@lines);
    }

    while (defined($line))
    {
        # adjust line number
        if ($reading_file)
        {
            $line_nr = $.;  # line number in file handle
        }
        else
        {
            $line_nr++;
        }

        chomp($line);

        # parse to line starting with `#-'
        last if ($line =~ /^\#-/);

        if ($line =~ /^\#\@(\w+)\s+(.*?)\s*$/)
        {
            my $value = $2;

            $key = $1;

            if ($value ne '')
            {
                if (exists($run_hash{$key}) && $run_hash{$key} ne '')
                {
                    $run_hash{$key} .= " $value";
                }
                else
                {
                    $run_hash{$key} = $value;
                }
            }
            else
            {
                mywarn("Warning: missing value to keyword `$key' in line ".
                     $line_nr);

                $run_hash{$key} = '' unless (exists($run_hash{$key}));
            }
        }
        elsif ($line =~ /^\#\s*(.*?)\s*$/ && defined($key))
        {
            my $value = $1;

            if ($value ne '')
            {
                if (exists($run_hash{$key}) && $run_hash{$key} ne '')
                {
                    $run_hash{$key} .= " $value";
                }
                else
                {
                    $run_hash{$key} = $value;
                }
            }
            else
            {
                mywarn("Warning: missing value to keyword `$key' in line ".
                     $line_nr);

                $run_hash{$key} = '' unless (exists($run_hash{$key}));
            }
        }

        # read new line
        if (defined($file_hdl))
        {
            $line = <$file_hdl>;
        }
        else
        {
            $line = shift(@lines);
        }
    }

    return %run_hash;
}
#### eo parse_run_info

####
# create_run_info_list
#
# Creates a list of run information based on the hash produced by the parsing
# subroutine.
#
# PARAMS:
#  \% - hash with parsed run information
# RETURNS:
#   @ - list of ordered run information (empty when error occurred)
#       (Machine, Institution, Nodes, Date (unchanged), Node, Network,
#        SKaMPI_User, SKaMPI_Version, OS_Name, OS_Release, OS_Version,
#        Hostname, Architecture, Absolute (0 or 1), Postproc, Comment)
#     (Postproc and Comment can be undef; Comment is the ` || '-separated
#      string of `key: value' strings for sorted other keywords)
# EXCEPTIONS:
#   * `MACHINE' definition is missing ...
#   * `INSTITUTION' definition is missing ...
#   * `NODES' definition is missing ...
#   * `NODES' definition is not a number ...
#   * `DATE' definition is missing ...
#   * `NODE' definition is missing ...
#   * `NETWORK' definition is missing ...
#   * `USER' definition is missing ...
#   * `SKAMPIVERSION' definition is missing ...
#   * `OSNAME' definition is missing ...
#   * `OSRELEASE' definition is missing ...
#   * `OSVERSION' definition is missing ...
#   * `HOSTNAME' definition is missing ...
#   * `ARCHITECTURE' definition is missing ...
#   * `ABSOLUTE' definition is missing ...
#   * `ABSOLUTE' definition must be `yes' or `no' ...
#   * `POSTPROC' definition is not a number ...
####
sub create_run_info_list
{
    my %run_hash  = %{shift;};
    my @info_list = ();

    foreach (qw(MACHINE INSTITUTION NODES DATE NODE NETWORK USER SKAMPIVERSION
                OSNAME OSRELEASE OSVERSION HOSTNAME ARCHITECTURE ABSOLUTE))
    {
        my $keyword = $_;

        unless (defined($run_hash{$keyword}) && $run_hash{$keyword} ne '')
        {
            quit("`$keyword' definition is missing");
            return ();
        }
        push(@info_list, $run_hash{$keyword});

        delete($run_hash{$keyword});
    }

    # INSTITUTION
    unless (check_institution($info_list[1]))
    {
        quit("`INSTITUTION' definition is missing");
        return ();
    }

    # NODES
    unless ($info_list[2] =~ /^\d+$/)
    {
        quit("`NODES' definition is not a number");
        return ();
    }

    # ABSOLUTE
    if ($info_list[13] eq 'yes')
    {
        $info_list[13] = 1;
    }
    elsif ($info_list[13] eq 'no')
    {
        $info_list[13] = 0;
    }
    else
    {
        quit("`ABSOLUTE' definition must be `yes' or `no'");
        return ();
    }

    # POSTPROC
    if (defined($run_hash{'POSTPROC'}))
    {
        unless ($run_hash{'POSTPROC'} =~ /^(\d+)\s+times$/)
        {
            quit("`POSTPROC' definition is not a number");
            return ();
        }
        push(@info_list, $1);

        delete($run_hash{'POSTPROC'});
    }
    else
    {
        push(@info_list, undef);
    }

    # other keywords as comments
    if (keys(%run_hash))
    {
        push(@info_list, join(' || ', map($run_hash{$_} ne '' ?
                                          "$_: $run_hash{$_}" : (),
                                          sort(keys(%run_hash)))));
    }
    else
    {
        push(@info_list, undef);
    }

    return @info_list;
}
#### eo create_run_info_list

####
# check_institution
#
# Checks if the `INSTITUTION' entry is defined and not one of the pseudo
# entries that SKaMPI produces.
#
# PARAMS:
#   $ - entry for the institution (can be undef)
# RETURNS:
#   $ - 0 if the entry is not valid (undef or pseudo string)
#       1 if the entry is valid
####
sub check_institution
{
    my $entry     = shift;
    my @def_insts = ('** please fill in the name and the location of the '.
                     'institution here**',   # SKaMPI versions < 3.0
                     '** please fill in the name and the location of your '.
                     'institution here **',  # SKaMPI versions >= 3.0
                     '?',  # produced by dorep2.pl      if entry is empty
                     '');  # produced by parse_run_info if entry is empty

    return 0 if (!defined($entry) || grep($entry eq $_, @def_insts));

    return 1;
}
#### eo check_institution

####
# parse_suite_info
#
# Parses the part of a SKaMPI output file where the information on a SKaMPI
# suite is located.
# If the parameter to this function is a file handle, the position within it
# after the function has ended is just before the measurements part.
#
# PARAMS:
#   @ - list of data lines
#       (if the first parameter is a file handle (or reference to it), read
#        from that one instead)
# RETURNS:
#   % - hash of suite information
#       (possible key/value pairs:
#        name          =>  $
#        pattern       =>  $
#        parameters    => \@
#        units         => \@
#        formats       => \@
#        xscales       => \@
#        refinements   => \@
#        rangestarts   => \@
#        rangeends     => \@
#        rangesteps    => \@
#        defaults      =>  $
#        stderror      =>  $
#        cutquantile   =>  $
#        resultformats =>  $)
#     (if the key of one of the lists exists the list's number of elements is
#      the same as that of the list of any other existing key)
# EXCEPTIONS:
#   * Dimension <NUM> instead of <NUM> in `Pattern' line <NUM> ...
#   * Wrong parameter description in `Pattern' line <NUM> ...
#   * Dimension <NUM> instead of <NUM> in `x scale' line <NUM> ...
#   * Dimension <NUM> instead of <NUM> in `range' line <NUM> at range
#       definition ...
#   * Dimension <NUM> instead of <NUM> in `range' line <NUM> at stepwidth
#       definition ...
#   * Wrong range description in `range' line <NUM> ...
####
sub parse_suite_info
{
    my @lines        = @_;
    my $file_hdl     = $lines[0];
    my %suite_hash   = ();
    my $dimension    = undef;  # dimension of suite (number of var. param.)
    my $reading_file = 0;      # parameters are data lines (no file handle)
    my $line_nr      = 0;      # number of parsed line (1-based)
    my $line;

    if (ref($lines[0]) eq 'GLOB' || ref(\$lines[0]) eq 'GLOB')
    {
        $reading_file = 1;

        $line = <$file_hdl>;
    }
    else
    {
        $line = shift(@lines);
    }

    while (defined($line))
    {
        # adjust line number
        if ($reading_file)
        {
            $line_nr = $.;  # line number in file handle
        }
        else
        {
            $line_nr++;
        }

        if ($line =~ /^\#Description of the (.+?) measurement:/)
        {
            $suite_hash{'name'} = $1;
        }
        elsif ($line =~ /^\#Pattern: (Simple)/)
        {
            $suite_hash{'pattern'} = $1;
        }
        elsif ($line =~ /^\#Pattern: (.+?) varied over the (.+)/)
        {
            my @params = split(/\s*;\s*/, $2);

            # look for inconsistencies of dimension of this suite
            if (defined($dimension) && $dimension != @params)
            {
                quit("Dimension ".scalar(@params)." instead of $dimension ".
                      "in `Pattern' line $line_nr");
            }
            $dimension = @params;

            $suite_hash{'pattern'}    = $1;
            $suite_hash{'parameters'} = [];
            $suite_hash{'units'}      = [];
            $suite_hash{'formats'}    = [];

            # parse description(s) of variation parameter(s)
            foreach (@params)
            {
                if (/(.+?) \[(.+?)\] \((\%.+)\)/)
                {
                    push(@{$suite_hash{'parameters'}}, $1);
                    push(@{$suite_hash{'units'}},      $2);
                    push(@{$suite_hash{'formats'}},    $3);
                }
                else
                {
                    quit("Wrong parameter description in `Pattern' line ".
                          $line_nr);
                }
            }
        }
        elsif (my @scales = parse_scale_line($line))
        {
            # look for inconsistencies of dimension of this suite
            if (defined($dimension) && $dimension != @scales)
            {
                quit("Dimension ".scalar(@scales)." instead of $dimension in".
                      " `x scale' line $line_nr");
            }
            $dimension = @scales;

            $suite_hash{'xscales'}     = [];
            $suite_hash{'refinements'} = [];

            # parse description(s) for scale and optional refinement
            foreach (@scales)
            {
                my ($scale, $refinement) = split(/\s*,\s*/, $_, 2);

                # refinement can be `undef' now
                push(@{$suite_hash{'xscales'}},     $scale);
                push(@{$suite_hash{'refinements'}}, $refinement);
            }
        }
        elsif ($line =~ /^\#range: (.+?), stepwidth: (.+?)\.$/)
        {
            my @ranges     = split(/\s*,\s*/, $1);
            my @stepwidths = split(/\s*,\s*/, $2);

            # look for inconsistencies of dimension of this suite
            if (defined($dimension) && $dimension != @ranges)
            {
                quit("Dimension ".scalar(@ranges)." instead of $dimension in".
                      " `range' line $line_nr at range definition");
            }
            $dimension = @ranges;

            if (defined($dimension) && $dimension != @stepwidths)
            {
                quit("Dimension ".scalar(@stepwidths)." instead of ".
                      "$dimension in `range' line $line_nr at stepwidth ".
                      "definition");
            }
            $dimension = @stepwidths;

            $suite_hash{'rangestarts'} = [];
            $suite_hash{'rangeends'}   = [];
            $suite_hash{'rangesteps'}  = [@stepwidths];

            # parse range definitions
            foreach (@ranges)
            {
                if (/(\d+) - (\d+)/)
                {
                    push(@{$suite_hash{'rangestarts'}}, $1);
                    push(@{$suite_hash{'rangeends'}},   $2);
                }
                else
                {
                    quit("Wrong range description in `range' line $line_nr");
                }
            }
        }
        elsif ($line =~ /^\#default values: (.*)/)
        {
            $suite_hash{'defaults'} = $1;
        }
        elsif ($line =~ /^\#max. allowed standard error is (\d+\.\d+) \%, cut quantile is (\d+\.\d+) \%/)
        {
            $suite_hash{'stderror'}    = $1;
            $suite_hash{'cutquantile'} = $2;
        }
        elsif ($line =~ /^\#Format:.*?<results>\s+(.*?)\s*$/)
        {
            $suite_hash{'resultformats'} = $1;

            # this has to be the last line of suite header information
            last;
        }

        # read new line
        if (defined($file_hdl))
        {
            $line = <$file_hdl>;
        }
        else
        {
            $line = shift(@lines);
        }
    }

    return %suite_hash;
}
#### eo parse_suite_info

####
# parse_scale_line
#
# Parses an `x scale' line of a SKaMPI output file where the information about
# the x scales of a SKaMPI suite is located.
# If the parameter to this function is a file handle, the position within it
# after the function has ended is just after the read line.
#
# PARAMS:
#   $ - data line or file handle
#       (if the first parameter is a file handle (or reference to it), read the
#        line to parse from that one instead)
# RETURNS:
#   @ - list of the scale descriptions per variation parameter
#       (empty if it's not an `x scale' line)
####
sub parse_scale_line
{
    my $line   = shift;
    my @scales = ();

    $line = <$line> if (ref($line) eq 'GLOB' || ref(\$line) eq 'GLOB');

    if ($line =~ /^\#The x scale is (.+)/)
    {
        @scales = split(/\s*;\s*/, $1);
    }

    return @scales;
}
#### eo parse_scale_line

####
# create_suite_info_list
#
# Creates a list of suite information based on the hash produced by the parsing
# subroutine.
#
# PARAMS:
#  \% - hash with parsed suite information
# RETURNS:
#   @ - list of ordered suite information (empty when error occurred)
#       (Name, Pattern, (Variable, ...), (Unit, ...), (Format, ...),
#        (X_Scale, ...), (Refinement, ...), (Range_Begin, ...),
#        (Range_End, ...), (Range_Step, ...), Default_Values, Std_Error,
#        Cut_Quantile, Results_Formats)
#     (each of the inner lists has the same number of elements - the dimension
#      of the suite - 0 with `Simple' pattern; each element in the Refinement
#      list can be undef; Default_Values can be undef)
# EXCEPTIONS:
#   * Definition of suite name is missing ...
#   * Definition of suite pattern is missing ...
#   * Definition. of ... ... missing ...
#     (with the second dots being any (comma, space)-separated non-empty subset
#      of (variation parameter(s), parameter units, parameter formats,
#      x scales, scale refinements, range start values, range end values,
#      range stepwidths) and the first and third dot(s) being plural 's' and
#      'are' or 'is' respectively)
#   * Definition of std. error is missing ...
#   * Definition of cut quantile is missing ...
#   * Definition of formats of results is missing ...
####
sub create_suite_info_list
{
    my %suite_hash = %{shift;};
    my @info_list  = ();
    my @undefined  = ();  # list of undefined list parameters

    foreach (qw(name pattern))
    {
        my $keyword = $_;

        unless (defined($suite_hash{$keyword}))
        {
            quit("Definition of suite $keyword is missing");
            return ();
        }
        push(@info_list, $suite_hash{$keyword});
    }

    foreach (['parameters', 'variation parameter(s)'],
             ['units', 'parameter units'], ['formats', 'parameter formats'],
             ['xscales', 'x scales'], ['refinements', 'scale refinements'],
             ['rangestarts', 'range start values'],
             ['rangeends', 'range end values'],
             ['rangesteps', 'range stepwidths'],
             ['defaults', 'default values'])
    {
        my $keyword   = $$_[0];
        my $param_str = $$_[1];

        if (defined($suite_hash{$keyword}) && $suite_hash{$keyword} !~ /^\s*$/)
        {
            push(@info_list, $suite_hash{$keyword});
        }
        else
        {
            if ($keyword eq 'defaults')
            {
                push(@info_list, undef);
            }
            else
            {
                push(@info_list, []);

                push(@undefined, $param_str);
            }
        }
    }

    # Either all of the list parameters may be defined or none at all
    if (@undefined != 0 && @undefined != 8)
    {
        my $plural = (@undefined != 1);

        quit(sprintf("Definition%s of %s %s missing",
                      $plural ? 's' : '', join(', ', @undefined),
                      $plural ? 'are' : 'is'));
    }

    foreach (['stderror', 'std. error'], ['cutquantile', 'cut quantile'],
             ['resultformats', 'formats of results'])
    {
        my $keyword   = $$_[0];
        my $param_str = $$_[1];

        unless (defined($suite_hash{$keyword}) && $suite_hash{$keyword} ne '')
        {
            quit("Definition of $param_str is missing");
            return ();
        }
        push(@info_list, $suite_hash{$keyword});
    }

    return @info_list;
}
#### eo create_suite_info_list

####
# parse_measurements
#
# Parses the part of a SKaMPI output file where the results of a suite's
# measurements are located.
# If the second parameter to this function is a file handle, the position
# within it after the function has ended is just after the measurements part
# (after the #eol-line).
#
# PARAMS:
#   $ - dimension of suite
#   @ - list of data lines
#       (if the first element is a file handle (or reference to it), read from
#        that one instead)
# RETURNS:
#   @ - list of the measurements' results
#       (((Value, ...), Result_Cl, Std_Error_Cl, Count_Cl, Result_All,
#         Std_Error_All, Count_All, (Node_Times)), ...)
#     (list of Node_Times can be empty; otherwise in descending order)
# EXCEPTIONS:
#   * Not enough numbers in measurement line <NUM> ...
####
sub parse_measurements
{
    my $dimension    = shift;  # dimension of suite (number of var. param.)
    my @lines        = @_;
    my $file_hdl     = $lines[0];
    my $reading_file = 0;      # parameters are data lines (no file handle)
    my $line_nr      = 0;      # number of parsed line (1-based)
    my @measurements = ();
    my @args_so_far  = ();     # to compare with new args (must be different)
    my $line;

    # adjust dimension (0 belongs to Simple pattern, but nevertheless there's
    # a pseudo parameter value to parse)
    $dimension = 1 unless ($dimension);

    if (ref($lines[0]) eq 'GLOB' || ref(\$lines[0]) eq 'GLOB')
    {
        $reading_file = 1;

        $line = <$file_hdl>;
    }
    else
    {
        $line = shift(@lines);
    }

    while (defined($line))
    {
        # adjust line number
        if ($reading_file)
        {
            $line_nr = $.;  # line number in file handle
        }
        else
        {
            $line_nr++;
        }

        # parse to line starting with #eol
        last if ($line =~ /^\#eol/);

        # treat lines with numbers starting with ## as normal result lines
        # (produced by dorep because of gnuplot problem; s. dorep)
        if (my @numbers = parse_measurement_line($line))
        {
            my @results    = ();  # results of a single measurement
            my $args_exist = 0;   # are the argument values already in suite?

            quit("Not enough numbers in measurement line $line_nr")
              if (@numbers < $dimension + 6);

            # argument list (number depending on dimension)
            push(@results, []);

            foreach (1..$dimension)
            {
                push(@{$results[0]}, shift(@numbers));
            }

            # check if an argument list with the same values already exists
            foreach (@args_so_far)
            {
                my @existing_args = @$_;
                my $eq_count      = 0;

                foreach (1..$dimension)
                {
                    $eq_count++ if ($existing_args[$_] == $ {$results[0]}[$_]);
                }

                if ($eq_count == $dimension)
                {
                    mywarn(sprintf("Warning: argument value%s in line $line_nr ".
                                 "already existing in the respective suite",
                                 $dimension == 1 ? '' : 's'));
                    $args_exist = 1;

                    last;
                }
            }

            next if ($args_exist);

            # list of result values
            foreach (1..6)
            {
                push(@results, shift(@numbers));
            }

            # optional sorted node times (@numbers can be empty by now)
            push(@results, [sort({$b <=> $a} @numbers)]);

            push(@measurements, \@results);
        }
    }
    continue
    {
        # read new line
        if ($reading_file)
        {
            $line = <$file_hdl>;
        }
        else
        {
            $line = shift(@lines);
        }
    }

    return @measurements;
}
#### eo parse_measurements

####
# parse_measurement_line
#
# Parses a measurement line of a SKaMPI output file.
# If the parameter to this function is a file handle, the position within it
# after the function has ended is just after the read line.
#
# PARAMS:
#   $ - data line or file handle
#       (if the first parameter is a file handle (or reference to it), read the
#        line to parse from that one instead)
# RETURNS:
#   @ - flat list of the numbers within the read line
#       (empty if it's not a measurement line)
####
sub parse_measurement_line
{
    my $line    = shift;
    my @numbers = ();

    $line = <$line> if (ref($line) eq 'GLOB' || ref(\$line) eq 'GLOB');

    if ($line =~ /^(?:\#\#)?\s*(\d+\s+.*?)\s*$/)
    {
        @numbers = split(/\s+/, $1);
    }

    return @numbers;
}
#### eo parse_measurement_line

########## end of functions ##########


__END__


# detailed listing of the collective suites and their respective bandwidth
# functions

# MPI_Allgather-length         ->  comp_bdw_col_all
# MPI_Allgather-nodes-long     ->  comp_bdw_col_all
# MPI_Allgather-nodes-short    ->  comp_bdw_col_all
# MPI_Allgatherv-length        ->  comp_bdw_col_all
# MPI_Allgatherv-nodes-long    ->  comp_bdw_col_all
# MPI_Allgatherv-nodes-short   ->  comp_bdw_col_all
# MPI_Allreduce-length         ->  comp_bdw_col_all
# MPI_Allreduce-nodes          ->  comp_bdw_col_all
# MPI_Alltoall-length          ->  comp_bdw_col_all
# MPI_Alltoall-nodes-long      ->  comp_bdw_col_all
# MPI_Alltoall-nodes-short     ->  comp_bdw_col_all
# MPI_Alltoallv-length         ->  comp_bdw_col_all
# MPI_Alltoallv-nodes-long     ->  comp_bdw_col_all
# MPI_Alltoallv-nodes-short    ->  comp_bdw_col_all
# MPI_Barrier-nodes            ->  undef (bandwidth useless)
# MPI_Bcast-length             ->  comp_bdw_col_sng
# MPI_Bcast-nodes-long         ->  comp_bdw_col_sng
# MPI_Bcast-nodes-short        ->  comp_bdw_col_sng
# MPI_Commsplit-nodes          ->  undef (bandwidth useless)
# MPI_Gather-length            ->  comp_bdw_col_sng
# MPI_Gather-nodes-long        ->  comp_bdw_col_sng
# MPI_Gather-nodes-short       ->  comp_bdw_col_sng
# MPI_Gather_ISWA-length       ->  comp_bdw_col_sng
# MPI_Gather_ISWA-nodes-long   ->  comp_bdw_col_sng
# MPI_Gather_ISWA-nodes-short  ->  comp_bdw_col_sng
# MPI_Gather_SR-length         ->  comp_bdw_col_sng
# MPI_Gather_SR-nodes-long     ->  comp_bdw_col_sng
# MPI_Gather_SR-nodes-short    ->  comp_bdw_col_sng
# MPI_Gatherv-length           ->  comp_bdw_col_sng
# MPI_Gatherv-nodes-long       ->  comp_bdw_col_sng
# MPI_Gatherv-nodes-short      ->  comp_bdw_col_sng
# MPI_Reduce-length            ->  comp_bdw_col_sng
# MPI_Reduce-nodes             ->  comp_bdw_col_sng
# MPI_Reduce_Bcast-length      ->  comp_bdw_col_sng
# MPI_Reduce_Bcast-nodes       ->  comp_bdw_col_sng
# MPI_Reduce_Scatterv-length   ->  comp_bdw_col_sng
# MPI_Reduce_Scatterv-nodes    ->  comp_bdw_col_sng
# MPI_Reduce_scatter-length    ->  comp_bdw_col_sng
# MPI_Reduce_scatter-nodes     ->  comp_bdw_col_sng
# MPI_Scan-length              ->  comp_bdw_col_sng
# MPI_Scan-nodes               ->  comp_bdw_col_sng
# MPI_Scatter-length           ->  comp_bdw_col_sng
# MPI_Scatter-nodes-long       ->  comp_bdw_col_sng
# MPI_Scatter-nodes-short      ->  comp_bdw_col_sng
# MPI_Scatterv-length          ->  comp_bdw_col_sng
# MPI_Scatterv-nodes-long      ->  comp_bdw_col_sng
# MPI_Scatterv-nodes-short     ->  comp_bdw_col_sng
# col_dummy                    ->  undef (bandwidth useless)
